# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
from sqlalchemy.orm import exc

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.api.rpc.handlers import l3_rpc
from neutron.api.v2 import attributes
from neutron.common import constants as l3_constants
from neutron.common import exceptions as q_exc
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.db import common_db_mixin
from neutron.db import extraroute_db
from neutron.db import l3_agentschedulers_db
from neutron.db import l3_db
from neutron.db import l3_dvr_db
from neutron.db import l3_gwmode_db
from neutron.db import models_v2
from neutron.extensions import l3
from neutron.extensions import portsecurity as psec
from neutron.i18n import _LE
from neutron.plugins.common import constants

from networking_brocade.vyatta.common import config
from networking_brocade.vyatta.common import exceptions as v_exc
from networking_brocade.vyatta.common import utils as vyatta_utils
from networking_brocade.vyatta.vrouter import driver as vrouter_driver


LOG = logging.getLogger(__name__)


class VyattaVRouterMixin(common_db_mixin.CommonDbMixin,
                         extraroute_db.ExtraRoute_db_mixin,
                         l3_dvr_db.L3_NAT_with_dvr_db_mixin,
                         l3_gwmode_db.L3_NAT_db_mixin,
                         l3_agentschedulers_db.L3AgentSchedulerDbMixin):
    """Brocade Neutron L3 Plugin for Vyatta vRouter.

    Supports CRUD operations on vRouter, add/remove interfaces from vRouter
    and floating IPs for VMs.It performs vRouter VM lifecyle management by
    calling Nova APIs during the Create and Delete Router calls.
    Once the vRouter VM is up, L3 plugin uses REST API to perform the
    configurations. L3 plugin supports add/remove router interfaces by
    attaching the neutron ports to vRouter VM using Nova API.
    RPC notifications will be used by the firewall agent that is coupled
    with l3-agent. This is needed for our firewall plugin.
    """

    ATTACH_PORT_RETRY_LIMIT = 5
    ATTACH_PORT_RETRY_DELAY = 5

    def __init__(self):
        self.setup_rpc()
        self.driver = vrouter_driver.VyattaVRouterDriver()
        self.router_scheduler = importutils.import_object(
            config.CONF.router_scheduler_driver)
        self.start_periodic_l3_agent_status_check()

    def setup_rpc(self):
        # RPC support
        self.topic = topics.L3PLUGIN
        self.conn = n_rpc.create_connection(new=True)
        self.agent_notifiers.update(
            {l3_constants.AGENT_TYPE_L3: l3_rpc_agent_api.L3AgentNotifyAPI()})
        self.endpoints = [_VyattaL3RPCEndpoint()]
        self.conn.create_consumer(self.topic, self.endpoints,
                                  fanout=False)
        self.conn.consume_in_threads()

    def get_plugin_type(self):
        return constants.L3_ROUTER_NAT

    def get_plugin_description(self):
        """Returns string description of the plugin."""
        return ("Brocade Vyatta Router Service Plugin for basic L3 forwarding "
                "between (L2) Neutron networks and access to external "
                "networks via a NAT gateway.")

    def create_router(self, context, router):
        """Creates the vRouter VM using vrouter_driver.

        If we encounter vRouter VM creation failure or connectivity failure
        vrouter_driver will handle the appropriate exceptions and delete
        the vRouter VM.
        """
        LOG.debug("Vyatta vRouter Plugin::Create router: %s", router)

        r = router['router']
        router_id = self.driver.create_router(context)
        if router_id is None:
            raise q_exc.BadRequest(
                resource='router',
                msg=_('Vyatta vRouter creation failed'))

        gw_info = r.pop(l3.EXTERNAL_GW_INFO, attributes.ATTR_NOT_SPECIFIED)

        tenant_id = self._get_tenant_id_for_create(context, r)

        with context.session.begin(subtransactions=True):
            # noinspection PyArgumentList
            router_db = l3_db.Router(id=router_id,
                                     tenant_id=tenant_id,
                                     name=r['name'],
                                     admin_state_up=r['admin_state_up'],
                                     status="ACTIVE")
            context.session.add(router_db)
            self._process_extra_attr_router_create(context, router_db, router)
            router_dict = self._make_router_dict(router_db)

        try:
            self.driver.init_router(context, router_dict)
        except (v_exc.InvalidVRouterInstance,
                v_exc.InvalidInstanceConfiguration,
                v_exc.VRouterConnectFailure,
                v_exc.VRouterOperationError,
                Exception):
            with excutils.save_and_reraise_exception():
                with context.session.begin(subtransactions=True):
                    context.session.delete(router_db)

        if gw_info != attributes.ATTR_NOT_SPECIFIED:
            self._update_router_gw_info(context, router_db['id'], gw_info)
            router_dict[l3.EXTERNAL_GW_INFO] = gw_info

        return self._make_router_dict(router_db)

    def update_router(self, context, router_id, router):
        LOG.debug("Vyatta vRouter Plugin::Update router: %s", router)

        r = router['router']

        gw_info = r.pop(l3.EXTERNAL_GW_INFO, attributes.ATTR_NOT_SPECIFIED)

        if gw_info != attributes.ATTR_NOT_SPECIFIED:
            self._update_router_gw_info(context, router_id, gw_info)

        return super(VyattaVRouterMixin, self).update_router(
            context, router_id, router)

    def delete_router(self, context, router_id):
        LOG.debug("Vyatta vRouter Plugin::Delete router: %s", router_id)

        gw_port = None

        with context.session.begin(subtransactions=True):
            router = self._get_router(context, router_id)

            self._ensure_router_not_in_use(context, router_id)

            # delete any gw port
            device_filter = {
                'device_id': [router_id],
                'device_owner': [l3_constants.DEVICE_OWNER_ROUTER_GW]
            }
            ports = self._core_plugin.get_ports(context.elevated(),
                                                filters=device_filter)

            if ports:
                gw_port = ports[0]
                router.gw_port = None
                context.session.add(router)

        if gw_port:
            self._delete_router_port(context, router_id, gw_port)

        with context.session.begin(subtransactions=True):
            context.session.delete(router)

        self.driver.delete_router(context, router_id)

        self.l3_rpc_notifier.router_deleted(context, router_id)

    def add_router_interface(self, context, router_id, interface_info):
        LOG.debug("Vyatta vRouter Plugin::Add Router Interface. "
                  "router: %s; interface: %s", router_id, interface_info)
        add_by_port, add_by_sub = self._validate_interface_info(interface_info)
        router = self._get_router(context, router_id)

        new_port = True

        if add_by_port:
            port, subnets = self._add_interface_by_port(
                    context, router, interface_info['port_id'], '')
        elif add_by_sub:
            port, subnets, new_port = self._add_interface_by_subnet(
                    context, router, interface_info['subnet_id'], '')

        if new_port:
            port_tenant_id = port['tenant_id']
            self._core_plugin._delete_port_security_group_bindings(
                context.elevated(), port['id'])
            port = self._core_plugin.update_port(
                context.elevated(), port['id'], {'port': {
                    'tenant_id': config.VROUTER.tenant_id,
                    'device_id': '',
                    psec.PORTSECURITY: False,
                }})

            try:
                self._attach_port(context, router_id, port)
            except Exception:
                with excutils.save_and_reraise_exception():
                    if add_by_sub:
                        try:
                            self._core_plugin.delete_port(context.elevated(),
                                                          port['id'])
                        except Exception:
                            LOG.exception(_LE(
                                'Failed to delete previously created '
                                'port for Vyatta vRouter.'))

            port = self._core_plugin.update_port(
                context.elevated(), port['id'], {'port': {
                    'tenant_id': port_tenant_id,
                }})

            with context.session.begin(subtransactions=True):
                router_port = l3_db.RouterPort(
                    port_id=port['id'],
                    router_id=router.id,
                    port_type=port['device_owner']
                )
                context.session.add(router_port)
        else:
            self._update_port(context, router_id, port)

        router_interface_info = self._make_router_interface_info(
            router_id, port['tenant_id'], port['id'], subnets[-1]['id'],
            [subnet['id'] for subnet in subnets])
        self.notify_router_interface_action(
            context, router_interface_info, 'add')
        return router_interface_info

    def remove_router_interface(self, context, router_id, interface_info):
        LOG.debug("Vyatta vRouter Plugin::Remove Router Interface. "
                  "router: %s; interface_info: %s", router_id, interface_info)

        remove_by_port, remove_by_subnet = (
            self._validate_interface_info(interface_info, for_removal=True)
        )
        port_id = interface_info.get('port_id')
        subnet_id = interface_info.get('subnet_id')
        router = self._get_router(context, router_id)
        device_owner = self._get_device_owner(context, router)

        if remove_by_port:
            port, subnets, delete_port = self._remove_interface_by_port(
                    context, router_id, port_id, subnet_id, device_owner)
        # remove_by_subnet is not used here, because the validation logic of
        # _validate_interface_info ensures that at least one of remote_by_*
        # is True.
        else:
            port, subnets, delete_port = self._remove_interface_by_subnet(
                    context, router_id, subnet_id, device_owner)

        port_tenant_id = port['tenant_id']
        if delete_port:
            port = self._core_plugin.update_port(
                context.elevated(), port['id'], {'port': {
                    'tenant_id': config.VROUTER.tenant_id,
                }})
            self._delete_router_port(context, router_id, port)
        else:
            self._update_port(context, router_id, port)

        router_interface_info = self._make_router_interface_info(
            router_id, port_tenant_id, port['id'], subnets[0]['id'],
            [subnet['id'] for subnet in subnets])
        self.notify_router_interface_action(
            context, router_interface_info, 'remove')
        return router_interface_info

    def _remove_interface_by_port(self, context, router_id,
                                  port_id, subnet_id, owner):
        qry = context.session.query(l3_db.RouterPort)
        qry = qry.filter_by(
            port_id=port_id,
            router_id=router_id,
            port_type=owner
        )
        try:
            port_db = qry.one().port
        except exc.NoResultFound:
            raise l3.RouterInterfaceNotFound(router_id=router_id,
                                             port_id=port_id)
        port_subnet_ids = [fixed_ip['subnet_id']
                           for fixed_ip in port_db['fixed_ips']]
        if subnet_id and subnet_id not in port_subnet_ids:
            raise q_exc.SubnetMismatchForPort(
                port_id=port_id, subnet_id=subnet_id)
        subnets = [self._core_plugin._get_subnet(context, port_subnet_id)
                   for port_subnet_id in port_subnet_ids]
        for port_subnet_id in port_subnet_ids:
            self._confirm_router_interface_not_in_use(
                    context, router_id, port_subnet_id)
        return port_db, subnets, True

    def _remove_interface_by_subnet(self, context,
                                    router_id, subnet_id, owner):
        self._confirm_router_interface_not_in_use(
            context, router_id, subnet_id)
        subnet = self._core_plugin._get_subnet(context, subnet_id)

        try:
            rport_qry = context.session.query(models_v2.Port).join(
                l3_db.RouterPort)
            ports = rport_qry.filter(
                l3_db.RouterPort.router_id == router_id,
                l3_db.RouterPort.port_type == owner,
                models_v2.Port.network_id == subnet['network_id']
            )

            for p in ports:
                port_subnets = [fip['subnet_id'] for fip in p['fixed_ips']]
                if subnet_id in port_subnets and len(port_subnets) > 1:
                    # multiple prefix port - delete prefix from port
                    fixed_ips = [fip for fip in p['fixed_ips'] if
                            fip['subnet_id'] != subnet_id]
                    p = self._core_plugin.update_port(context, p['id'],
                            {'port':
                                {'fixed_ips': fixed_ips}})
                    return p, [subnet], False
                elif subnet_id in port_subnets:
                    return p, [subnet], True
        except exc.NoResultFound:
            pass
        raise l3.RouterInterfaceNotFoundForSubnet(router_id=router_id,
                                                  subnet_id=subnet_id)

    def _get_interface_infos(self, context, port):
        LOG.debug("Vyatta vRouter Plugin::Get interface infos")

        mac_address = port['mac_address']
        interface_infos = []
        for fip in port['fixed_ips']:
            try:
                subnet = self._core_plugin._get_subnet(context.elevated(),
                                                       fip['subnet_id'])
                ipnet = netaddr.IPNetwork(subnet.cidr)
                interface_infos.append({
                    'mac_address': mac_address,
                    'ip_address': '{0}/{1}'.format(fip['ip_address'],
                                                   ipnet.prefixlen),
                    'gateway_ip': subnet.gateway_ip
                })
            except q_exc.SubnetNotFound:
                pass
        return interface_infos

    def _get_interface_infos_new(self, context, port):
        mac_address = port['mac_address']
        ip_list = []

        for fixed_ip in port['fixed_ips']:
            subnet = self._core_plugin._get_subnet(
                context.elevated(), fixed_ip['subnet_id'])
            network = netaddr.IPNetwork(subnet.cidr)
            ip = netaddr.IPNetwork('{0}/{1}'.format(
                fixed_ip['ip_address'], network.prefixlen))
            ip_list.append(ip)

        return vyatta_utils.InterfaceInfo(
            ip_addresses=ip_list, mac_address=mac_address)

    def _delete_router_port(self, context, router_id, port, external_gw=False):
        # Get instance, deconfigure interface and detach port from it. To do
        # this need to change port owner back to that instance.
        LOG.debug("Vyatta vRouter Plugin::Delete router port. "
                  "router: %s; port: %s", router_id, port)

        self.driver.deconfigure_interface(
            context, router_id, self._get_interface_infos(context.elevated(),
                                                          port))

        self._core_plugin.update_port(context.elevated(), port['id'],
                                      {'port': {'device_owner': '',
                                                'device_id': router_id}})
        self.driver.detach_interface(context, router_id, port['id'])

        self._core_plugin.delete_port(context.elevated(), port['id'])

    def _attach_port(self, context, router_id, port, external_gw=False):
        LOG.debug("Vyatta vRouter Plugin::Attach port. "
                  "router: %s; port: %s", router_id, port)
        # Attach interface
        self.driver.attach_interface(context, router_id, port['id'])

        def configure_gateway_wrapper():
            if external_gw:
                self.driver.configure_gateway(
                    context, router_id,
                    self._get_interface_infos(context, port))
            else:
                self.driver.configure_interface(
                    context, router_id,
                    self._get_interface_infos(context, port))

        vyatta_utils.retry(
            configure_gateway_wrapper,
            exceptions=(v_exc.VRouterOperationError,),
            limit=self.ATTACH_PORT_RETRY_LIMIT,
            delay=self.ATTACH_PORT_RETRY_DELAY)

        if external_gw:
            device_owner = l3_constants.DEVICE_OWNER_ROUTER_GW
        else:
            device_owner = l3_constants.DEVICE_OWNER_ROUTER_INTF
        self._core_plugin.update_port(context.elevated(), port['id'],
                                      {'port': {'device_owner': device_owner,
                                                'device_id': router_id}})

    def _update_port(self, context, router_id, port):
        interface_info = self._get_interface_infos_new(context, port)
        self.driver.update_interface(context, router_id, interface_info)

    def _update_router_gw_info(self, context, router_id, info, router=None):
        LOG.debug("Vyatta vRouter Plugin::Update router gateway info")

        router = router or self._get_router(context, router_id)
        gw_port = router.gw_port
        ext_ips = info.get('external_fixed_ips') if info else []
        network_id = self._validate_gw_info(context, gw_port, info, ext_ips)
        ext_ip_change = self._check_for_external_ip_change(
            context, gw_port, ext_ips)

        if gw_port and ext_ip_change and gw_port['network_id'] == network_id:
            self._update_current_gw_port(context, router_id, router,
                                         ext_ips)
        else:
            self._delete_current_gw_port(context, router_id, router,
                                         network_id)
            self._create_gw_port(context, router_id, router, network_id,
                                 ext_ips)

    def _update_current_gw_port(self, context, router_id, router, ext_ips):
        super(VyattaVRouterMixin, self)._update_current_gw_port(
            context, router_id, router, ext_ips)

        port = router.gw_port

        self.driver.clear_gateway(context, router_id)
        self.driver.configure_gateway(
            context, router_id,
            self._get_interface_infos(context.elevated(), port))

    def _delete_current_gw_port(self, context, router_id, router, new_network):
        """Delete gw port if attached to an old network or IPs changed."""
        port_requires_deletion = (
            router.gw_port and router.gw_port['network_id'] != new_network)
        if not port_requires_deletion:
            return
        admin_ctx = context.elevated()
        if self.get_floatingips_count(
                admin_ctx, {'router_id': [router_id]}):
            raise l3.RouterExternalGatewayInUseByFloatingIp(
                router_id=router_id, net_id=router.gw_port['network_id'])

        gw_port = router.gw_port
        self.driver.clear_gateway(context, router_id)
        with context.session.begin(subtransactions=True):
            router.gw_port = None
            context.session.add(router)
            context.session.expire(gw_port)
            self._check_router_gw_port_in_use(context, router_id)

        self._delete_router_port(
            context, router_id, gw_port, external_gw=True)

    def _create_router_gw_port(self, context, router, network_id, ext_ips):
        if ext_ips and len(ext_ips) > 1:
            msg = _("Routers support only 1 external IP")
            raise q_exc.BadRequest(resource='router', msg=msg)

        gw_port = self._core_plugin.create_port(context.elevated(), {
            'port': {
                'tenant_id': config.VROUTER.tenant_id,
                'network_id': network_id,
                'mac_address': attributes.ATTR_NOT_SPECIFIED,
                'fixed_ips': ext_ips or attributes.ATTR_NOT_SPECIFIED,
                'device_owner': '',
                'device_id': '',
                'admin_state_up': True,
                'name': '',
                psec.PORTSECURITY: False,
            }})

        if not gw_port['fixed_ips']:
            self._core_plugin.delete_port(context.elevated(), gw_port['id'],
                                          l3_port_check=False)
            msg = (_('No IPs available for external network %s') %
                   network_id)
            raise q_exc.BadRequest(resource='router', msg=msg)

        with context.session.begin(subtransactions=True):
            router.gw_port = self._core_plugin._get_port(context.elevated(),
                                                         gw_port['id'])
            router_port = l3_db.RouterPort(
                router_id=router.id,
                port_id=gw_port['id'],
                port_type=l3_constants.DEVICE_OWNER_ROUTER_GW
            )
            context.session.add(router)
            context.session.add(router_port)

        try:
            self._attach_port(context, router['id'], gw_port,
                              external_gw=True)
        except Exception as ex:
            LOG.exception(_LE("Exception while attaching port : %s"), ex)
            with excutils.save_and_reraise_exception():
                try:
                    with context.session.begin(subtransactions=True):
                        router.gw_port = None
                        context.session.add(router)
                        self._core_plugin.delete_port(context.elevated(),
                                                      gw_port['id'])
                except Exception:
                    LOG.exception(_LE('Failed to roll back changes to '
                                    'Vyatta vRouter after external '
                                    'gateway assignment.'))

    def _update_extra_routes(self, context, router, routes):
        LOG.debug(
            'Vyatta vRouter Plugin::update static routes. '
            'router_id={0}'.format(router['id']))

        routes_old = self._get_extra_routes_by_router_id(
            context, router['id'])
        super(VyattaVRouterMixin, self)._update_extra_routes(
            context, router, routes)
        routes_new = self._get_extra_routes_by_router_id(
            context, router['id'])

        routes_old = self._route_rules_to_set(routes_old)
        routes_new = self._route_rules_to_set(routes_new)

        self.driver.update_static_routes(
            context, router['id'],
            tuple(routes_new - routes_old),
            tuple(routes_old - routes_new))

    @staticmethod
    def _route_rules_to_set(rules):
        result = set()
        for r in rules:
            result.add(vyatta_utils.RouteRule(
                dest_cidr=r['destination'], next_hop=r['nexthop']))
        return result

    def create_floatingip(
            self, context, floatingip,
            initial_status=l3_constants.FLOATINGIP_STATUS_ACTIVE):
        LOG.debug("Vyatta vRouter Plugin::Create floating ip")

        floatingip_dict = super(VyattaVRouterMixin, self).create_floatingip(
            context, floatingip,
            initial_status=initial_status)
        router_id = floatingip_dict['router_id']
        if router_id:
            self.associate_floatingip(context, router_id, floatingip_dict)
        return floatingip_dict

    def associate_floatingip(self, context, router_id, floatingip):
        LOG.debug("Vyatta vRouter Plugin::Associate floating ip")

        fixed_ip = floatingip['fixed_ip_address']
        floating_ip = floatingip['floating_ip_address']
        if router_id:
            self.driver.assign_floating_ip(
                context, router_id, floating_ip, fixed_ip)
            with context.session.begin(subtransactions=True):
                floatingip_db = self._get_floatingip(context, floatingip['id'])
                floatingip_db['status'] = l3_constants.FLOATINGIP_STATUS_ACTIVE

    def update_floatingip(self, context, floatingip_id, floatingip):
        LOG.debug("Vyatta vRouter Plugin::Update floating ip")

        fip = floatingip['floatingip']
        with context.session.begin(subtransactions=True):
            floatingip_db = self._get_floatingip(context, floatingip_id)
            old_floatingip = self._make_floatingip_dict(floatingip_db)
            fip['tenant_id'] = floatingip_db['tenant_id']
            fip['id'] = floatingip_id
            fip_port_id = floatingip_db['floating_port_id']
            before_router_id = floatingip_db['router_id']
            self._update_fip_assoc(context, fip, floatingip_db,
                                   self._core_plugin.get_port(
                                       context.elevated(), fip_port_id))

        if before_router_id:
            self.disassociate_floatingip(
                context, before_router_id, old_floatingip)

        router_id = floatingip_db['router_id']
        if router_id:
            self.associate_floatingip(context, router_id, floatingip_db)

        return self._make_floatingip_dict(floatingip_db)

    def delete_floatingip(self, context, floatingip_id):
        LOG.debug("Vyatta vRouter Plugin::Delete floating ip: %s",
                  floatingip_id)

        floatingip_dict = self._get_floatingip(context, floatingip_id)
        router_id = floatingip_dict['router_id']
        if router_id:
            self.disassociate_floatingip(context, router_id, floatingip_dict)
        super(VyattaVRouterMixin, self).delete_floatingip(
            context, floatingip_id)

    def disassociate_floatingip(self, context, router_id, floatingip):
        LOG.debug("Vyatta vRouter Plugin::Disassociate floating ip."
                  "router: %s; floating_ip: %s", router_id, floatingip)

        fixed_ip = floatingip['fixed_ip_address']
        floating_ip = floatingip['floating_ip_address']
        if router_id:
            self.driver.unassign_floating_ip(
                context, router_id, floating_ip, fixed_ip)
            with context.session.begin(subtransactions=True):
                floatingip_db = self._get_floatingip(context, floatingip['id'])
                floatingip_db['status'] = l3_constants.FLOATINGIP_STATUS_DOWN

    def disassociate_floatingips(self, context, port_id, do_notify=True):
        LOG.debug("Vyatta vRouter Plugin::Disassociate floating ips."
                  "port_id: %s", port_id)

        with context.session.begin(subtransactions=True):
            fip_qry = context.session.query(l3_db.FloatingIP)
            floating_ips = fip_qry.filter_by(fixed_port_id=port_id)
            for floating_ip in floating_ips:
                self.disassociate_floatingip(
                    context, floating_ip['router_id'], floating_ip)

        return super(VyattaVRouterMixin, self).disassociate_floatingips(
            context, port_id, do_notify)


class _VyattaL3RPCEndpoint(l3_rpc.L3RpcCallback):
    def sync_routers(self, context, **kwargs):
        routers_list = super(_VyattaL3RPCEndpoint, self).sync_routers(
            context, **kwargs)

        if not routers_list:
            return routers_list

        routers_by_id = dict((x['id'], x) for x in routers_list)

        query = context.session.query(models_v2.Port)
        query = query.filter(models_v2.Port.network_id
                             == config.VROUTER.management_network_id)
        query = query.filter(models_v2.Port.device_id.in_(routers_by_id))

        need_processed = set(routers_by_id)
        for port in query:
            router_id = port['device_id']

            try:
                need_processed.remove(router_id)
            except KeyError:
                raise v_exc.CorruptedSystemError(
                    description=(
                        'router {0} contain multiple interface joined to '
                        'management network').format(router_id))

            # this statement can't raise KeyError because query condition
            router = routers_by_id[router_id]

            try:
                ip = port['fixed_ips']
                ip = ip[0]
                ip = ip['ip_address']
            except (IndexError, KeyError):
                raise v_exc.CorruptedSystemError(
                    description=(
                        'vyatta vrouter id={0} management interface have no '
                        'ip address').format(router_id))

            router['_vyatta'] = {
                'management_ip_address': ip}

        if need_processed:
            raise v_exc.CorruptedSystemError(
                description=(
                    'vyatta vrouters not linked to management network: '
                    '{0}').format(', '.join(sorted(need_processed))))

        return routers_list



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Implementation of SSH Connector"""

import networking_brocade.mlx.ml2.commands as Commands
from networking_brocade.mlx.ml2.device_connector import (
    DeviceConnector as DevConn)
from neutron.i18n import _LE
from oslo_log import log as logging
import paramiko

LOG = logging.getLogger(__name__)
WRITE = "wb"
READ = "rb"


class SSHConnector(DevConn):

    """
    Uses SSH to connect to device
    """

    def connect(self):
        """
        Connect to the device
        """
        try:
            self.connector = paramiko.SSHClient()
            self.connector.set_missing_host_key_policy(
                paramiko.AutoAddPolicy())
            self.connector.connect(
                hostname=self.host,
                username=self.username,
                password=self.password)

            channel = self.connector.invoke_shell()
            self.stdin_stream = channel.makefile(WRITE)
            self.stdout_stream = channel.makefile(READ)
            self.stderr_stream = channel.makefile(READ)

        except Exception as e:
            LOG.exception(_LE("Connect failed to switch %(host)s with error"
                              " %(error)s"),
                          {'host': self.host, 'error': e.args})
            raise Exception(_("Connection Failed"))

    def write(self, command):
        """
        Write from input stream to device

        :param:command: Command to be executed on the device
        """
        self.stdin_stream.write(command)
        self.stdin_stream.flush()

    def read_response(self, read_lines=True):
        """Read the response from the output stream.

        :param:read_lines: Boolean value which indicated to read multiple line
            or single line. It is true by default.
        :returns: Response from the device as list when read_lines is True or
            string when read_lines is false.
        """
        response = None
        if read_lines:
            response = self.stdout_stream.readlines()
        else:
            response = self.stdout_stream.read()
        return response

    def send_exit(self, count):
        """Send Exit command.

        :param:count: Indicates number of times to execute exit command
        """
        index = 0
        while index < count:
            self.stdin_stream.write(Commands.EXIT)
            self.stdin_stream.flush()
            index += 1

    def close_session(self):
        """Close SSH session."""
        if self.connector:
            self.stdin_stream.close()
            self.stdout_stream.close()
            self.stderr_stream.close()
            self.connector.close()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from networking_brocade.vdx.ml2driver import (
    mechanism_brocade as brocademechanism)
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit.plugins.ml2 import test_plugin
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)

MECHANISM_NAME = ('networking_brocade.'
                  'vdx.ml2driver.mechanism_brocade.BrocadeMechanism')


class TestBrocadeMechDriverV2(test_plugin.Ml2PluginV2TestCase):
    """Test Brocade VCS/VDX mechanism driver.

    """

    _mechanism_name = MECHANISM_NAME

    def setUp(self):

        _mechanism_name = MECHANISM_NAME

        ml2_opts = {
            'mechanism_drivers': ['brocade'],
            'tenant_network_types': ['vlan']}

        for opt, val in ml2_opts.items():
            ml2_config.cfg.CONF.set_override(opt, val, 'ml2')

        def mocked_brocade_init(self):
            self._driver = mock.MagicMock()

        with mock.patch.object(brocademechanism.BrocadeMechanism,
                               'brocade_init', new=mocked_brocade_init):
            super(TestBrocadeMechDriverV2, self).setUp()
            self.mechanism_driver = importutils.import_object(_mechanism_name)


class TestBrocadeMechDriverNetworksV2(test_plugin.TestMl2NetworksV2,
                                      TestBrocadeMechDriverV2):
    pass


class TestBrocadeMechDriverPortsV2(test_plugin.TestMl2PortsV2,
                                   TestBrocadeMechDriverV2):
    pass


class TestBrocadeMechDriverSubnetsV2(test_plugin.TestMl2SubnetsV2,
                                     TestBrocadeMechDriverV2):
    pass


class TestBrocadeMechDriverFeaturesEnabledTestCase(TestBrocadeMechDriverV2):

    def setUp(self):
        super(TestBrocadeMechDriverFeaturesEnabledTestCase, self).setUp()

    def test_version_features(self):

        vf = True
        # Test for NOS version 4.0.3
        self.mechanism_driver.set_features_enabled("4.0.3", vf)
        # Verify
        pp_domain_support, virtual_fabric_enabled = (
            self.mechanism_driver.get_features_enabled()
        )
        self.assertFalse(pp_domain_support)
        self.assertTrue(virtual_fabric_enabled)

        # Test for NOS version 4.1.0
        vf = True
        self.mechanism_driver.set_features_enabled("4.1.0", vf)
        # Verify
        pp_domain_support, virtual_fabric_enabled = (
            self.mechanism_driver.get_features_enabled()
        )
        self.assertTrue(pp_domain_support)
        self.assertTrue(virtual_fabric_enabled)

        # Test for NOS version 4.1.3
        vf = False
        self.mechanism_driver.set_features_enabled("4.1.3", vf)
        # Verify
        pp_domain_support, virtual_fabric_enabled = (
            self.mechanism_driver.get_features_enabled()
        )
        self.assertTrue(pp_domain_support)
        self.assertFalse(virtual_fabric_enabled)

        # Test for NOS version 5.0.0
        vf = True
        self.mechanism_driver.set_features_enabled("5.0.0", vf)
        # Verify
        pp_domain_support, virtual_fabric_enabled = (
            self.mechanism_driver.get_features_enabled()
        )
        self.assertTrue(pp_domain_support)
        self.assertTrue(virtual_fabric_enabled)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import itertools
import logging
import urllib
import uuid

from oslo_utils import excutils
import six

from neutron.i18n import _LE, _LI
from networking_brocade.vyatta.common import exceptions as v_exc
from networking_brocade.vyatta.common import parsers as v_parsers
from networking_brocade.vyatta.vrouter import client as vyatta_client


LOG = logging.getLogger(__name__)

_KEY_CONNECTIONS = 'ipsec_site_connections'
_KEY_IKEPOLICY = 'ikepolicy'
_KEY_ESPPOLICY = 'ipsecpolicy'


def connect_setup_commands(vrouter, iface, svc, conn, resources):
        LOG.info(_LI('Vyatta vRouter: _connect_setup_commands:  '))
        SCmd = vyatta_client.SetCmd
        batch = list()

        if resources.iface_alloc(conn, iface):
            batch.append(
                SCmd('vpn/ipsec/ipsec-interfaces/interface/{0}'.format(iface)))

        ike_name, need_create = resources.ike_group_alloc(conn)
        if need_create:
            batch.extend(ike_setup_commands(conn, ike_name))

        esp_name, need_create = resources.esp_group_alloc(conn)
        if need_create:
            batch.extend(esp_setup_commands(conn, esp_name))

        link_id = [uuid.UUID(x).get_hex() for x in svc['id'], conn['id']]
        link_id.insert(0, 'os-id')
        link_id = ':'.join(link_id)
        remote_peer = conn['peer_address']
        p = 'vpn/ipsec/site-to-site/peer/{0}'.format(remote_peer)
        batch.extend([
            SCmd('{0}/description/{1}'.format(p, link_id)),
            SCmd('{0}/authentication/mode/pre-shared-secret'.format(p)),
            SCmd('{0}/authentication/pre-shared-secret/{1}'.format(
                p, conn['psk'])),
            SCmd('{0}/ike-group/{1}'.format(p, ike_name)),
            SCmd('{0}/default-esp-group/{1}'.format(p, esp_name)),
            SCmd('{0}/local-address/{1}'.format(
                p, urllib.quote_plus(svc['external_ip'])))])
        for remote_cidr in conn['peer_cidrs']:
            idx = resources.tunnel_alloc(conn, remote_peer)
            batch.append(SCmd(
                '{0}/tunnel/{1}/allow-public-networks/enable'.format(p, idx)))
            batch.append(SCmd('{0}/tunnel/{1}/local/prefix/{2}'.format(
                p, idx, urllib.quote_plus(svc['subnet']['cidr']))))
            batch.append(SCmd('{0}/tunnel/{1}/remote/prefix/{2}'.format(
                p, idx, urllib.quote_plus(remote_cidr))))
            LOG.info(_LI('Vyatta vRouter: _connect_setup_commands: '
                         'add snat-exclude for remote_cidr %s'), remote_cidr)
            vrouter.add_snat_exclude_rule(batch, iface,
                svc['subnet']['cidr'], remote_cidr)

        # TODO(dbogun): static routing for remote networks
        return batch


def ike_setup_commands(conn, name):
        policy = conn[_KEY_IKEPOLICY]

        SCmd = vyatta_client.SetCmd

        ike_prefix = 'vpn/ipsec/ike-group/{0}'.format(urllib.quote_plus(name))
        return [
            SCmd('{0}/proposal/1'.format(ike_prefix)),
            SCmd('{0}/proposal/1/encryption/{1}'.format(
                ike_prefix, policy['encryption_algorithm'])),
            SCmd('{0}/proposal/1/hash/{1}'.format(
                ike_prefix, policy['auth_algorithm'])),
            SCmd('{0}/lifetime/{1}'.format(
                ike_prefix, policy['lifetime_value'])),
            SCmd('{0}/dead-peer-detection/action/{1}'.format(
                ike_prefix, conn['dpd_action'])),
            SCmd('{0}/dead-peer-detection/interval/{1}'.format(
                ike_prefix, conn['dpd_interval'])),
            SCmd('{0}/dead-peer-detection/timeout/{1}'.format(
                ike_prefix, conn['dpd_timeout']))
        ]


def esp_setup_commands(conn, name):
        policy = conn[_KEY_ESPPOLICY]

        SCmd = vyatta_client.SetCmd

        esp_prefix = 'vpn/ipsec/esp-group/{0}'.format(urllib.quote_plus(name))
        return [
            SCmd('{0}/proposal/1'.format(esp_prefix)),
            SCmd('{0}/proposal/1/encryption/{1}'.format(
                esp_prefix, policy['encryption_algorithm'])),
            SCmd('{0}/proposal/1/hash/{1}'.format(
                esp_prefix, policy['auth_algorithm'])),
            SCmd('{0}/lifetime/{1}'.format(
                esp_prefix, policy['lifetime_value'])),
            SCmd('{0}/pfs/{1}'.format(
                esp_prefix, policy['pfs'])),
            SCmd('{0}/mode/{1}'.format(
                esp_prefix, policy['encapsulation_mode']))
        ]


def connect_remove_commands(vrouter, iface, svc, conn, resources):
        LOG.info(_LI('Vyatta vRouter: _connect_setup_commands:  '))

        DCmd = vyatta_client.DeleteCmd

        batch = [
            DCmd('vpn/ipsec/site-to-site/peer/{0}'.format(
                conn['peer_address']))]

        name, need_remove = resources.ike_group_release(conn)
        if need_remove:
            batch.append(
                DCmd('vpn/ipsec/ike-group/{0}'.format(
                    urllib.quote_plus(name))))
        name, need_remove = resources.esp_group_release(conn)
        if need_remove:
            batch.append(
                DCmd('vpn/ipsec/esp-group/{0}'.format(
                    urllib.quote_plus(name))))

        if resources.iface_release(conn, iface):
            # FIXME(dbogun): vrouter failed to complete this command
            # batch.append(
            #     DCmd('vpn/ipsec/ipsec-interfaces/interface/{0}'.format(
            #         iface)))
            pass

        for remote_cidr in conn['peer_cidrs']:
            LOG.info(_LI('Vyatta vRouter: _connect_setup_commands: '
                         'delete snat-exclude for remote_cidr %s'),
                     remote_cidr)
            vrouter.delete_snat_exclude_rule(batch, iface,
                svc['subnet']['cidr'], remote_cidr)

        resources.tunnel_release(conn, svc['external_ip'])
        return batch


def compare_vpn_services(vrouter, gw_iface, old, new):
    if old.get('_reversed', False):
        old_conn_by_id = dict((x['id'], x) for x in old[_KEY_CONNECTIONS])
        for conn_new in new[_KEY_CONNECTIONS]:
            try:
                old_conn = old_conn_by_id[conn_new['id']]
            except KeyError:
                continue
            old_conn['psk'] = conn_new['psk']

    batch_old = set()
    batch_new = set()
    for svc, batch in (
            (old, batch_old), (new, batch_new)):
        for conn in svc[_KEY_CONNECTIONS]:
            patch = RouterResources(
                svc['id']).make_patch()
            commands = connect_setup_commands(
                vrouter, gw_iface, svc, conn, patch)
            commands = tuple((x.make_url('_dummy_') for x in commands))
            batch.add(commands)
    return batch_old == batch_new


VPN_STATE_MAP = {
    'up': True,
    'down': False
}


def parse_vpn_connections(ipsec_sa, resources):
    parser = v_parsers.TableParser()
    parser(ipsec_sa)
    assert not (len(parser) % 2)

    conn_status = collections.defaultdict(list)
    parser_iter = iter(parser)
    for endpoints, tunnels in itertools.izip(parser_iter, parser_iter):
        endpoints = iter(endpoints)
        try:
            peer = next(endpoints)
            peer = peer.cell_by_idx(0)
        except StopIteration:
            raise ValueError('Invalid VPN IPSec status report.')

        for tunn in tunnels:
            tunn_idx = tunn.cell_by_name('Tunnel')
            try:
                tunn_idx = int(tunn_idx)
            except ValueError:
                raise v_exc.InvalidResponseFormat(
                    details=('expect integer on place of tunnel index '
                             'got {!r}').format(tunn_idx))

            try:
                conn_id = resources.get_connect_by_tunnel(
                    peer, tunn_idx)
            except v_exc.ResourceNotFound:
                continue

            state = tunn.cell_by_name('State')
            state = state.lower()
            try:
                state = VPN_STATE_MAP[state]
            except KeyError:
                raise v_exc.InvalidResponseFormat(
                    details='unsupported tunnel state {!r}'.format(
                        state))
            conn_status[conn_id].append(state)

    conn_ok = set(key for key, value in six.iteritems(conn_status)
                  if all(value))

    return conn_ok


def validate_svc_connection(conn):
    # NOTE(asaprykin): Maybe it's better to move code into
    # separate function to avoid long try..except blocks
    try:
        map_encryption = {
            '3des': '3des',
            'aes-128': 'aes128',
            'aes-256': 'aes256'}
        map_pfs = {
            'group2': 'dh-group2',
            'group5': 'dh-group5'}

        allowed_pfs = map_pfs.values() + ['enable', 'disable']

        if conn['dpd_action'] not in ('hold', 'clear', 'restart'):
            raise ValueError('invalid dpd_action {0}'.format(
                conn['dpd_action']))

        ike_policy = conn[_KEY_IKEPOLICY]
        ike_policy['encryption_algorithm'] = \
            map_encryption[ike_policy['encryption_algorithm']]
        if ike_policy['lifetime_units'] != 'seconds':
            raise ValueError(
                'invalid "lifetime_units"=="{}" in ike_policy'.format(
                    ike_policy['lifetime_units']))

        esp_policy = conn[_KEY_ESPPOLICY]
        esp_policy['encryption_algorithm'] = \
            map_encryption[esp_policy['encryption_algorithm']]
        if esp_policy['lifetime_units'] != 'seconds':
            raise ValueError(
                'invalid "lifetime_units"=="{}" in esp_policy'.format(
                    esp_policy['lifetime_units']))
        if esp_policy['transform_protocol'] != 'esp':
            raise ValueError(
                'invalid "transform_protocol"=="{}" in esp_policy'.format(
                    esp_policy['transform_protocol']))
        pfs = esp_policy['pfs']
        esp_policy['pfs'] = map_pfs.get(pfs, pfs)
        if esp_policy['pfs'] not in allowed_pfs:
            raise ValueError(
                'invalid "pfs"=="{}" in esp_policy'.format(pfs))
        if esp_policy['encapsulation_mode'] not in ('tunnel', 'transport'):
            raise ValueError(
                'invalid "encapsulation_mode"=="{}" in esp_policy'.format(
                    esp_policy['encapsulation_mode']))
    except (ValueError, KeyError):
        with excutils.save_and_reraise_exception():
            raise v_exc.InvalidVPNServiceError()


# -- Tools ------------------------
class RouterResources(object):
    def __init__(self, router_id):
        self.router_id = router_id

        self.iface_to_conn = collections.defaultdict(set)
        self.ike_to_conn = collections.defaultdict(set)
        self.esp_to_conn = collections.defaultdict(set)
        self.conn_to_tunn = collections.defaultdict(set)
        self.peer_to_conn = dict()
        self.tunn_idx_factory = collections.defaultdict(itertools.count)

    def make_patch(self):
        return _RouteResourcePatch(self)

    def key_for_conn(self, conn):
        return ('connect', conn['id'])

    def key_for_tunnel(self, conn, remote_peer):
        key = list(self.key_for_conn(conn))
        key[0] = 'tunn'
        key.append(remote_peer)
        return tuple(key)

    def get_connect_by_tunnel(self, remote_peer, idx):
        key = (remote_peer, idx)
        try:
            conn_id = self.peer_to_conn[key]
        except KeyError:
            raise v_exc.ResourceNotFound
        return conn_id


class _RouteResourcePatch(object):
    NAME_LENGTH_LIMIT = 251

    def __init__(self, owner):
        self._owner = owner
        self._actions = list()
        self._tunn_factory_overlay = dict()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            return

        self._apply()

    def iface_alloc(self, conn, name):
        key = self._owner.key_for_conn(conn)
        self._actions.append(_PatchSetAdd(
            self._owner.iface_to_conn[key], conn['id']))
        return not len(self._owner.iface_to_conn[key])

    def iface_release(self, conn, name):
        key = self._owner.key_for_conn(conn)
        self._actions.append(_PatchSetDel(
            self._owner.iface_to_conn[key], conn['id']))
        return self._owner.iface_to_conn[key] == {conn['id']}

    def ike_group_alloc(self, conn):
        return self._group_alloc(self._owner.ike_to_conn, conn, _KEY_IKEPOLICY)

    def esp_group_alloc(self, conn):
        return self._group_alloc(self._owner.esp_to_conn, conn, _KEY_ESPPOLICY)

    def _group_alloc(self, target, conn, policy_key):
        name = self._make_entity_name(conn[policy_key])
        self._actions.append(_PatchSetAdd(target[name], conn['id']))
        return name, not len(target[name])

    def ike_group_release(self, conn):
        return self._group_release(
            self._owner.ike_to_conn, conn, _KEY_IKEPOLICY)

    def esp_group_release(self, conn):
        return self._group_release(
            self._owner.esp_to_conn, conn, _KEY_ESPPOLICY)

    def _group_release(self, target, conn, policy_key):
        name = self._make_entity_name(conn[policy_key])
        self._actions.append(_PatchSetDel(target[name], conn['id']))
        return name, target[name] == {conn['id']}

    def tunnel_alloc(self, conn, remote_peer, idx=None):
        conn_key = self._owner.key_for_conn(conn)
        tunn_key = self._owner.key_for_tunnel(conn, remote_peer)
        if idx is None:
            idx = next(self._owner.tunn_idx_factory[tunn_key])
        else:
            try:
                max_idx, overlay = self._tunn_factory_overlay[tunn_key]
                max_idx = max(idx, max_idx)
                overlay.value = itertools.count(max_idx + 1)
            except KeyError:
                max_idx = idx
                overlay = _PatchDictAdd(
                    self._owner.tunn_idx_factory, tunn_key,
                    itertools.count(max_idx + 1))
                self._actions.append(overlay)
            self._tunn_factory_overlay[tunn_key] = max_idx, overlay

        self._actions.append(_PatchDictAdd(
            self._owner.peer_to_conn, (remote_peer, idx), conn['id']))
        self._actions.append(_PatchSetAdd(
            self._owner.conn_to_tunn[conn_key], idx))
        return idx

    def tunnel_release(self, conn, remote_peer, idx=None):
        key = self._owner.key_for_conn(conn)

        if idx is None:
            idx_seq = self._owner.conn_to_tunn[key]
        else:
            idx_seq = (idx, )

        for idx in idx_seq:
            self._actions.append(_PatchDictDel(
                self._owner.peer_to_conn, (remote_peer, idx)))
            self._actions.append(_PatchSetDel(
                self._owner.conn_to_tunn[key], idx))

    def _apply(self):
        for action in self._actions:
            action()

    def _make_entity_name(self, data):
        idnr = uuid.UUID(data['id'])
        name = data['name'].lower()
        name = ''.join((x if x.isalnum() else '') for x in name)
        name = '{0}-{1}'.format(name, idnr.get_hex())
        if self.NAME_LENGTH_LIMIT < len(name):
            raise v_exc.InvalidParameter(
                cause=('Can\'t make vyatta resource identifier, result exceed '
                       'length limit'))
        return name


@six.add_metaclass(abc.ABCMeta)
class _PatchActionAbstract(object):
    def __init__(self, target):
        self.target = target

    @abc.abstractmethod
    def __call__(self):
        pass


class _PatchDictAdd(_PatchActionAbstract):
    def __init__(self, target, key, value):
        _PatchActionAbstract.__init__(self, target)
        self.key = key
        self.value = value

    def __call__(self):
        self.target[self.key] = self.value


class _PatchDictDel(_PatchActionAbstract):
    def __init__(self, target, key, allow_missing=True):
        _PatchActionAbstract.__init__(self, target)
        self.key = key
        self.allow_missing = allow_missing

    def __call__(self):
        try:
            del self.target[self.key]
        except KeyError:
            if not self.allow_missing:
                raise v_exc.ResourceNotFound


class _PatchSetAdd(_PatchActionAbstract):
    def __init__(self, target, value):
        _PatchActionAbstract.__init__(self, target)
        self.value = value

    def __call__(self):
        self.target.add(self.value)


class _PatchSetDel(_PatchActionAbstract):
    def __init__(self, target, value, allow_missing=True):
        _PatchActionAbstract.__init__(self, target)
        self.value = value
        self.allow_missing = allow_missing

    def __call__(self):
        try:
            self.target.remove(self.value)
        except KeyError:
            if not self.allow_missing:
                raise v_exc.ResourceNotFound


def parse_vrouter_config(config, resources):
    try:
        config = config['vpn']
        config = config['ipsec']
        interfaces = config['ipsec-interfaces']
        site_to_site = config['site-to-site']
    except KeyError:
        return tuple()

    interfaces = interfaces.values()
    svc_set = collections.defaultdict(dict)

    key_prefix = 'peer '
    for peer in site_to_site:
        conn = site_to_site[peer]
        if not peer.startswith(key_prefix):
            continue
        peer = peer[len(key_prefix):]

        try:
            conn_data, svc_data = _parse_ipsec_site_to_site(
                peer, conn, config, resources)
        except v_exc.InvalidResponseFormat as e:
            LOG.error(
                _LE('process vRouter ipsec configuration: {0}').format(e))
            continue

        svc_id = svc_data['id']
        svc = svc_set[svc_id]
        svc.setdefault('_reversed', True)
        svc.update(svc_data)
        svc.setdefault('ipsec_site_connections', list()).append(conn_data)

        resources.ike_group_alloc(conn_data)
        resources.esp_group_alloc(conn_data)
        for iface in interfaces:
            resources.iface_alloc(conn_data, iface)
        for idx, remote in conn_data.pop('_tunn_and_cidr'):
            resources.tunnel_alloc(conn_data, remote, idx)

    return svc_set.values()


def _parse_ipsec_site_to_site(peer, conn, config, resources):
    result = dict()
    svc_upd = dict()

    try:
        idnr_set = conn['description'].split(':')
        if idnr_set.pop(0) != 'os-id':
            raise ValueError

        svc_id, conn_id = [str(uuid.UUID(x)) for x in idnr_set]
    except (KeyError, TypeError, ValueError):
        raise v_exc.InvalidResponseFormat(
            details='vpn connection does not contain neutron connection id')

    try:
        auth = conn['authentication']

        result['id'] = conn_id
        result['peer_address'] = peer
        result['psk'] = auth['pre-shared-secret']

        svc_upd['id'] = svc_id
        svc_upd['external_ip'] = conn['local-address']
    except KeyError:
        raise v_exc.InvalidResponseFormat(
            details='incomplete connection config')

    svc_upd['subnet'] = subnet = dict()

    key_prefix = 'tunnel '
    peer_cidr = list()
    for key in (x for x in conn if x.startswith(key_prefix)):
        idx, local, remote = _parse_ipsec_tunnel(key, conn)
        peer_cidr.append((idx, remote))
        subnet['cidr'] = local
    peer_cidr.sort(key=lambda x: x[0])
    result['_tunn_and_cidr'] = peer_cidr
    result['peer_cidrs'] = [x[1] for x in peer_cidr]

    ike, upd = _parse_ipsec_ike_group(conn, config, resources)
    result['ikepolicy'] = ike
    result.update(upd)

    esp, upd = _parse_ipsec_esp_group(conn, config, resources)
    result['ipsecpolicy'] = esp
    result.update(upd)

    return result, svc_upd


def _parse_ipsec_ike_group(conn, config, resources):
    result = dict()
    conn_upd = dict()

    try:
        name = conn['ike-group']
        policy = config['ike-group ' + name]
    except KeyError:
        raise v_exc.InvalidResponseFormat(details='ike group missing')

    name, idnr = _unpack_ipsec_group_name(name)

    try:
        dpd = policy['dead-peer-detection']
        proposal = policy['proposal 1']

        result['id'] = idnr
        result['name'] = name
        result['encryption_algorithm'] = proposal['encryption']
        result['auth_algorithm'] = proposal['hash']
        result['lifetime_value'] = policy['lifetime']

        conn_upd['dpd_action'] = dpd['action']
        conn_upd['dpd_interval'] = dpd['interval']
        conn_upd['dpd_timeout'] = dpd['timeout']
    except KeyError:
        raise v_exc.InvalidResponseFormat(
            details='incomplete IKE group config')

    return result, conn_upd


def _parse_ipsec_esp_group(conn, config, resources):
    result = dict()
    try:
        name = conn['default-esp-group']
        policy = config['esp-group ' + name]
    except KeyError:
        raise v_exc.InvalidResponseFormat(details='eps group missing')

    name, idnr = _unpack_ipsec_group_name(name)

    try:
        proposal = policy['proposal 1']

        result['id'] = idnr
        result['name'] = name
        result['encryption_algorithm'] = proposal['encryption']
        result['auth_algorithm'] = proposal['hash']
        result['lifetime_value'] = policy['lifetime']
        result['pfs'] = policy['pfs']
        result['encapsulation_mode'] = policy['mode']
    except KeyError:
        raise v_exc.InvalidResponseFormat(
            details='incomplete ESP group config')

    return result, dict()


def _parse_ipsec_tunnel(key, conn):
    try:
        peer = conn[key]
        idx = key.rsplit(' ', 1)[1]
        idx = int(idx)
    except (KeyError, IndexError, ValueError):
        raise v_exc.InvalidResponseFormat(
            details='invalid tunnel section "{0}"'.format(key))

    try:
        local = peer['local']
        local = local['prefix']

        remote = peer['remote']
        remote = remote['prefix']
    except KeyError:
        raise v_exc.InvalidResponseFormat(
            details='incomplete peer config')

    return idx, local, remote


def _unpack_ipsec_group_name(raw):
    try:
        name, idnr = raw.rsplit('-', 1)
        idnr = uuid.UUID(idnr)
        idnr = str(idnr)
    except (TypeError, ValueError):
        raise v_exc.InvalidResponseFormat(
            details='can\'t parse group name "{0}"'.format(raw))
    return name, idnr



# Copyright 2014 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Brocade NOS Driver implements NETCONF over SSHv2 for
Neutron network life-cycle management.
"""

from ncclient import manager
from oslo_log import log as logging
from oslo_utils import excutils

from xml.etree import ElementTree

from networking_brocade.vdx.ml2driver.nos import nctemplates as template
from neutron.i18n import _LE

LOG = logging.getLogger(__name__)
SSH_PORT = 22


def nos_unknown_host_cb(host, fingerprint):
    """An unknown host callback.

    Returns `True` if it finds the key acceptable,
    and `False` if not. This default callback for NOS always returns 'True'
    (i.e. trusts all hosts for now).
    """
    return True


class NOSdriver(object):
    """NOS NETCONF interface driver for Neutron network.

    Handles life-cycle management of Neutron network (leverages AMPP on NOS)
    """

    def __init__(self):
        self.mgr = None
        self._virtual_fabric_enabled = False
        self._pp_domains_supported = False

    def set_features_enabled(self, pp_domains_supported,
                             virtual_fabric_enabled):
        """Set features in the driver based on what was detected by the MD."""
        self._pp_domains_supported = pp_domains_supported
        self._virtual_fabric_enabled = virtual_fabric_enabled

    def get_features_enabled(self):
        """Respond to status of features enabled."""
        return self._pp_domains_supported, self._virtual_fabric_enabled

    def connect(self, host, username, password):
        """Connect via SSH and initialize the NETCONF session."""

        # Use the persisted NETCONF connection
        if self.mgr and self.mgr.connected:
            return self.mgr

        # check if someone forgot to edit the conf file with real values
        if host == '':
            raise Exception(_("Brocade Switch IP address is not set, "
                              "check config ml2_conf_brocade.ini file"))

        # Open new NETCONF connection
        try:
            self.mgr = manager.connect(host=host, port=SSH_PORT,
                                       username=username, password=password,
                                       unknown_host_cb=nos_unknown_host_cb)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("Connect failed to switch"))

        LOG.debug("Connect success to host %(host)s:%(ssh_port)d",
                  dict(host=host, ssh_port=SSH_PORT))
        return self.mgr

    def close_session(self):
        """Close NETCONF session."""
        if self.mgr:
            self.mgr.close_session()
            self.mgr = None

    def get_nos_version(self, host, username, password):
        """Show version of NOS."""
        try:
            mgr = self.connect(host, username, password)
            return self.nos_version_request(mgr)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def is_virtual_fabric_enabled(self, host, username, password):
        """Show version of NOS."""
        try:
            mgr = self.connect(host, username, password)
            return (self.virtual_fabric_info(mgr) == "enabled")
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def create_network(self, host, username, password, net_id):
        """Creates a new virtual network."""

        domain_name = "default"
        name = template.OS_PORT_PROFILE_NAME.format(id=net_id)
        try:
            mgr = self.connect(host, username, password)
            self.create_vlan_interface(mgr, net_id)
            self.create_port_profile(mgr, name)

            if self._pp_domains_supported and self._virtual_fabric_enabled:
                self.configure_port_profile_in_domain(mgr, domain_name, name)

            self.create_vlan_profile_for_port_profile(mgr, name)

            if self._pp_domains_supported:
                self.configure_l2_mode_for_vlan_profile_with_domains(mgr, name)
            else:
                self.configure_l2_mode_for_vlan_profile(mgr, name)

            self.configure_trunk_mode_for_vlan_profile(mgr, name)
            self.configure_allowed_vlans_for_vlan_profile(mgr, name, net_id)
            self.activate_port_profile(mgr, name)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def delete_network(self, host, username, password, net_id):
        """Deletes a virtual network."""

        domain_name = "default"
        name = template.OS_PORT_PROFILE_NAME.format(id=net_id)
        try:
            mgr = self.connect(host, username, password)
            if self._pp_domains_supported and self._virtual_fabric_enabled:
                self.remove_port_profile_from_domain(mgr, domain_name, name)
            self.deactivate_port_profile(mgr, name)
            self.delete_port_profile(mgr, name)
            self.delete_vlan_interface(mgr, net_id)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def associate_mac_to_network(self, host, username, password,
                                 net_id, mac):
        """Associates a MAC address to virtual network."""

        name = template.OS_PORT_PROFILE_NAME.format(id=net_id)
        try:
            mgr = self.connect(host, username, password)
            self.associate_mac_to_port_profile(mgr, name, mac)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def dissociate_mac_from_network(self, host, username, password,
                                    net_id, mac):
        """Dissociates a MAC address from virtual network."""

        name = template.OS_PORT_PROFILE_NAME.format(id=net_id)
        try:
            mgr = self.connect(host, username, password)
            self.dissociate_mac_from_port_profile(mgr, name, mac)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def create_vlan_interface(self, mgr, vlan_id):
        """Configures a VLAN interface."""

        confstr = template.CREATE_VLAN_INTERFACE.format(vlan_id=vlan_id)
        mgr.edit_config(target='running', config=confstr)

    def delete_vlan_interface(self, mgr, vlan_id):
        """Deletes a VLAN interface."""

        confstr = template.DELETE_VLAN_INTERFACE.format(vlan_id=vlan_id)
        mgr.edit_config(target='running', config=confstr)

    def get_port_profiles(self, mgr):
        """Retrieves all port profiles."""

        filterstr = template.PORT_PROFILE_XPATH_FILTER
        response = mgr.get_config(source='running',
                                  filter=('xpath', filterstr)).data_xml
        return response

    def get_port_profile(self, mgr, name):
        """Retrieves a port profile."""

        filterstr = template.PORT_PROFILE_NAME_XPATH_FILTER.format(name=name)
        response = mgr.get_config(source='running',
                                  filter=('xpath', filterstr)).data_xml
        return response

    def create_port_profile(self, mgr, name):
        """Creates a port profile."""

        confstr = template.CREATE_PORT_PROFILE.format(name=name)
        mgr.edit_config(target='running', config=confstr)

    def delete_port_profile(self, mgr, name):
        """Deletes a port profile."""

        confstr = template.DELETE_PORT_PROFILE.format(name=name)
        mgr.edit_config(target='running', config=confstr)

    def activate_port_profile(self, mgr, name):
        """Activates a port profile."""

        confstr = template.ACTIVATE_PORT_PROFILE.format(name=name)
        mgr.edit_config(target='running', config=confstr)

    def deactivate_port_profile(self, mgr, name):
        """Deactivates a port profile."""

        confstr = template.DEACTIVATE_PORT_PROFILE.format(name=name)
        mgr.edit_config(target='running', config=confstr)

    def associate_mac_to_port_profile(self, mgr, name, mac_address):
        """Associates a MAC address to a port profile."""

        confstr = template.ASSOCIATE_MAC_TO_PORT_PROFILE.format(
            name=name, mac_address=mac_address)
        mgr.edit_config(target='running', config=confstr)

    def dissociate_mac_from_port_profile(self, mgr, name, mac_address):
        """Dissociates a MAC address from a port profile."""

        confstr = template.DISSOCIATE_MAC_FROM_PORT_PROFILE.format(
            name=name, mac_address=mac_address)
        mgr.edit_config(target='running', config=confstr)

    def create_vlan_profile_for_port_profile(self, mgr, name):
        """Creates VLAN sub-profile for port profile."""

        confstr = template.CREATE_VLAN_PROFILE_FOR_PORT_PROFILE.format(
            name=name)
        mgr.edit_config(target='running', config=confstr)

    def configure_l2_mode_for_vlan_profile(self, mgr, name):
        """Configures L2 mode for VLAN sub-profile."""

        confstr = template.CONFIGURE_L2_MODE_FOR_VLAN_PROFILE.format(
            name=name)
        mgr.edit_config(target='running', config=confstr)

    def configure_trunk_mode_for_vlan_profile(self, mgr, name):
        """Configures trunk mode for VLAN sub-profile."""

        confstr = template.CONFIGURE_TRUNK_MODE_FOR_VLAN_PROFILE.format(
            name=name)
        mgr.edit_config(target='running', config=confstr)

    def configure_allowed_vlans_for_vlan_profile(self, mgr, name, vlan_id):
        """Configures allowed VLANs for VLAN sub-profile."""

        confstr = template.CONFIGURE_ALLOWED_VLANS_FOR_VLAN_PROFILE.format(
            name=name, vlan_id=vlan_id)
        mgr.edit_config(target='running', config=confstr)

    def remove_port_profile_from_domain(self, mgr, domain_name, name):
        """Remove port-profile from default domain."""
        confstr = template.REMOVE_PORTPROFILE_FROM_DOMAIN.format(
            domain_name=domain_name, name=name)
        mgr.edit_config(target='running', config=confstr)

    def configure_port_profile_in_domain(self, mgr, domain_name, name):
        """put port-profile in default domain."""
        confstr = template.CONFIGURE_PORTPROFILE_IN_DOMAIN.format(
            domain_name=domain_name, name=name)
        mgr.edit_config(target='running', config=confstr)

    def configure_l2_mode_for_vlan_profile_with_domains(self, mgr, name):
        """Configures L2 mode for VLAN sub-profile."""
        confstr = template.CONFIGURE_L2_MODE_FOR_VLAN_PROFILE_IN_DOMAIN.format(
            name=name)
        mgr.edit_config(target='running', config=confstr)

    def nos_version_request(self, mgr):
        """Get firmware information using NETCONF rpc."""
        reply = mgr.dispatch(template.SHOW_FIRMWARE_VERSION, None, None)
        et = ElementTree.fromstring(str(reply))
        return et.find(template.NOS_VERSION).text

    def virtual_fabric_info(self, mgr):
        """Get virtual fabric info using NETCONF get-config."""
        response = mgr.get_config('running',
                                  filter=("xpath", "/vcs/virtual-fabric"))
        et = ElementTree.fromstring(str(response))
        vfab_enable = et.find(template.VFAB_ENABLE)
        if vfab_enable is not None:
            return "enabled"
        return "disabled"

    def create_svi(self, host, username, password,
                   rbridge_id, vlan_id, ip_address, router_id):
        """create svi on configured rbridge-id."""
        try:
            mgr = self.connect(host, username, password)
            self.bind_vrf_to_svi(host, username, password,
                                 rbridge_id, vlan_id, router_id)
            self.configure_svi_with_ip_address(mgr,
                                               rbridge_id, vlan_id, ip_address)
            self.activate_svi(mgr, rbridge_id, vlan_id)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error: %s"), ex)
                self.close_session()

    def delete_svi(self, host, username, password,
                   rbridge_id, vlan_id, gw_ip, router_id):
        """delete svi from configured rbridge-id."""
        try:
            mgr = self.connect(host, username, password)
            self.remove_svi(mgr, rbridge_id, vlan_id)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error: %s"), ex)
                self.close_session()

    def create_router(self, host, username, password, rbridge_id, router_id):
        """create vrf and associate vrf."""
        router_id = router_id[0:11]
        vrf_name = template.OS_VRF_NAME.format(id=router_id)
        rd = router_id + ":" + router_id
        try:
            mgr = self.connect(host, username, password)
            self.create_vrf(mgr, rbridge_id, vrf_name)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()
        try:
            # For Nos5.0.0
            self.configure_rd_for_vrf(mgr, rbridge_id, vrf_name, rd)
            self.configure_address_family_for_vrf(mgr, rbridge_id, vrf_name)
        except Exception:
            with excutils.save_and_reraise_exception() as ctxt:
                try:
                    # This is done because on 4.0.0 rd doesnt accept alpha
                    # character nor hyphen
                    rd = "".join(i for i in router_id if i in "0123456789")
                    rd = rd[:4] + ":" + rd[:4]
                    self.configure_rd_for_vrf(mgr, rbridge_id, vrf_name, rd)
                    self.configure_address_family_for_vrf_v1(mgr,
                                                             rbridge_id,
                                                             vrf_name)
                except Exception:
                    with excutils.save_and_reraise_exception():
                        LOG.exception(_LE("NETCONF error"))
                        self.close_session()

                ctxt.reraise = False

    def delete_router(self, host, username, password, rbridge_id, router_id):
        """delete router and associated vrf."""
        router_id = router_id[0:11]
        vrf_name = template.OS_VRF_NAME.format(id=router_id)
        try:
            mgr = self.connect(host, username, password)
            self.delete_vrf(mgr, rbridge_id, vrf_name)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def bind_vrf_to_svi(self, host, username, password, rbridge_id,
                        vlan_id, router_id):
        """binds vrf to a svi."""
        router_id = router_id[0:11]
        vrf_name = template.OS_VRF_NAME.format(id=router_id)
        try:
            mgr = self.connect(host, username, password)
            self.add_vrf_to_svi(mgr, rbridge_id, vlan_id, vrf_name)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def unbind_vrf_to_svi(self, host, username, password, rbridge_id,
                          vlan_id, router_id):
        """unbind vrf from the svi."""
        router_id = router_id[0:11]
        vrf_name = template.OS_VRF_NAME.format(id=router_id)
        try:
            mgr = self.connect(host, username, password)
            self.delete_vrf_from_svi(mgr, rbridge_id, vlan_id, vrf_name)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(_LE("NETCONF error"))
                self.close_session()

    def create_vrf(self, mgr, rbridge_id, vrf_name):
        """create vrf on rbridge."""
        confstr = template.CREATE_VRF.format(rbridge_id=rbridge_id,
                                             vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def delete_vrf(self, mgr, rbridge_id, vrf_name):
        """delete vrf on rbridge."""

        confstr = template.DELETE_VRF.format(rbridge_id=rbridge_id,
                                             vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def configure_rd_for_vrf(self, mgr, rbridge_id, vrf_name, rd):
        """configure rd on vrf  on rbridge."""

        confstr = template.CONFIGURE_RD_FOR_VRF.format(rbridge_id=rbridge_id,
                                                       vrf_name=vrf_name,
                                                       rd=rd)
        mgr.edit_config(target='running', config=confstr)

    def configure_address_family_for_vrf_v1(self, mgr, rbridge_id, vrf_name):
        """configure ipv4 address family to vrf  on rbridge."""

        confstr = template.ADD_ADDRESS_FAMILY_FOR_VRF_V1.format(
            rbridge_id=rbridge_id,
            vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def configure_address_family_for_vrf(self, mgr, rbridge_id, vrf_name):
        """configure ipv4 address family to vrf  on rbridge."""

        confstr = template.ADD_ADDRESS_FAMILY_FOR_VRF.format(
            rbridge_id=rbridge_id, vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def configure_svi_with_ip_address(self, mgr, rbridge_id,
                                      vlan_id, ip_address):
        """configure SVI with ip address on rbridge."""

        confstr = template.CONFIGURE_SVI_WITH_IP_ADDRESS.format(
            rbridge_id=rbridge_id,
            vlan_id=vlan_id,
            ip_address=ip_address)

        mgr.edit_config(target='running', config=confstr)

    def activate_svi(self, mgr, rbridge_id, vlan_id):
        """activate the svi on the rbridge."""
        confstr = template.ACTIVATE_SVI.format(rbridge_id=rbridge_id,
                                               vlan_id=vlan_id)
        mgr.edit_config(target='running', config=confstr)

    def add_vrf_to_svi(self, mgr, rbridge_id, vlan_id, vrf_name):
        """add vrf to svi on rbridge."""
        confstr = template.ADD_VRF_TO_SVI.format(rbridge_id=rbridge_id,
                                                 vlan_id=vlan_id,
                                                 vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def delete_vrf_from_svi(self, mgr, rbridge_id, vlan_id, vrf_name):
        """delete vrf from svi on rbridge."""
        confstr = template.DELETE_VRF_FROM_SVI.format(rbridge_id=rbridge_id,
                                                      vlan_id=vlan_id,
                                                      vrf_name=vrf_name)
        mgr.edit_config(target='running', config=confstr)

    def remove_svi(self, mgr, rbridge_id, vlan_id):
        """delete vrf from svi on rbridge."""
        confstr = template.DELETE_SVI.format(rbridge_id=rbridge_id,
                                             vlan_id=vlan_id)
        mgr.edit_config(target='running', config=confstr)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg

from neutron import context
from neutron.db import db_base_plugin_v2
from neutron.db import external_net_db
from neutron.db import models_v2
from neutron.db import securitygroups_rpc_base as sg_db_rpc
from neutron.extensions import l3
from neutron.openstack.common import uuidutils
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_db_plugin
from neutron.tests.unit.extensions import test_l3 as test_l3_plugin
from neutron.tests.unit import testlib_api


from networking_brocade.vyatta.common import utils as vyatta_utils
from networking_brocade.vyatta.vrouter import neutron_plugin as vrouter_plugin

_uuid = uuidutils.generate_uuid


class FakeVRouterDriver(mock.Mock):
    def create_router(self, *args, **kwargs):
        return _uuid()


class VRouterTestPlugin(vrouter_plugin.VyattaVRouterMixin,
                        db_base_plugin_v2.NeutronDbPluginV2,
                        external_net_db.External_net_db_mixin,
                        sg_db_rpc.SecurityGroupServerRpcMixin):

    def delete_port(self, context, port_id, l3_port_check=False):
        super(VRouterTestPlugin, self).delete_port(context, port_id)


class TestVyattaVRouterPlugin(testlib_api.SqlTestCase):
    def setUp(self):
        super(TestVyattaVRouterPlugin, self).setUp()

        self.setup_coreplugin(__name__ + '.' + VRouterTestPlugin.__name__)

        self._mock('eventlet.greenthread.sleep')

        self.driver = mock.Mock(wraps=FakeVRouterDriver())
        fake_driver_mock = mock.Mock()
        fake_driver_mock.return_value = self.driver

        self._mock(
            'networking_brocade.vyatta.vrouter.driver.VyattaVRouterDriver',
            fake_driver_mock)

        self.context = context.get_admin_context()
        self.plugin = VRouterTestPlugin()

        session = self.context.session
        with session.begin(subtransactions=True):
            self.ext_net = self._make_net('ext', is_external=True)
            self.ext_subnet = self._make_subnet(
                'ext', '10.10.10', self.ext_net['id'])
            self.ext_port = self._make_port('f0', self.ext_net['id'])
            self._make_fixed_ip(
                self.ext_port['id'], self.ext_net['id'],
                self.ext_subnet['id'], '10.10.10.22')

    def _mock(self, target, new=mock.DEFAULT):
        patcher = mock.patch(target, new)
        return patcher.start()

    def _mock_object(self, target, attribute, new=mock.DEFAULT):
        patcher = mock.patch.object(target, attribute, new)
        return patcher.start()

    def _make_net(self, n, is_shared=False, is_external=False):
        session = self.context.session
        network = models_v2.Network(tenant_id='fake-tenant-id',
                                    name='test-network-{0}'.format(n),
                                    status='ACTIVE',
                                    admin_state_up=True,
                                    shared=is_shared)
        session.add(network)
        session.flush()
        if is_external:
            extnet = external_net_db.ExternalNetwork(
                network_id=network['id'])
            session.add(extnet)
            session.flush()
        return network

    def _make_subnet(self, n, cidr_prefix, network_id):
        session = self.context.session
        subnet = models_v2.Subnet(tenant_id='fake-tenant-id',
                                  name='test-subnet-{0}'.format(n),
                                  network_id=network_id,
                                  ip_version=4,
                                  cidr='{0}.0/24'.format(cidr_prefix),
                                  gateway_ip='{0}.1'.format(cidr_prefix),
                                  enable_dhcp=True,
                                  shared=False)
        session.add(subnet)
        session.flush()
        ippool = models_v2.IPAllocationPool(
            subnet_id=subnet['id'],
            first_ip='{0}.1'.format(cidr_prefix),
            last_ip='{0}.254'.format(cidr_prefix))
        session.add(ippool)
        session.flush()
        iprange = models_v2.IPAvailabilityRange(
            allocation_pool_id=ippool['id'],
            first_ip='{0}.1'.format(cidr_prefix),
            last_ip='{0}.254'.format(cidr_prefix))
        session.add(iprange)
        session.flush()
        return subnet

    def _make_fixed_ip(self, port_id, network_id, subnet_id, ip):
        session = self.context.session
        ip_allocation = models_v2.IPAllocation(
            port_id=port_id,
            ip_address=ip,
            subnet_id=subnet_id,
            network_id=network_id)
        session.add(ip_allocation)
        session.flush()
        return ip_allocation

    def _make_port(self, port, network_id, device_id=None, device_owner=None):
        session = self.context.session
        port = models_v2.Port(tenant_id='fake-tenant-id',
                              name='',
                              network_id=network_id,
                              mac_address='aa:bb:cc:dd:ee:{0}'.format(port),
                              admin_state_up=True,
                              status='ACTIVE',
                              device_id=device_id or '',
                              device_owner=device_owner or '')
        session.add(port)
        session.flush()
        return port

    def test_create_router(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        result = self.plugin.create_router(self.context, router_data)
        self.assertTrue(uuidutils.is_uuid_like(result.get('id')))
        self.driver.create_router.assert_called_once_with(mock.ANY)

    def test_update_router1(self):
        router_data = {
            'router': {
                'name': 'test_router1',
                'admin_state_up': True,
                'external_gateway_info': {},
            }
        }
        router = self.plugin.create_router(self.context, router_data)
        router_new = self.plugin.update_router(self.context, router['id'], {
            'router': {
                'name': 'router2',
                'external_gateway_info': {},
            }
        })
        self.assertEqual(router_new['name'], 'router2')

    def test_update_router2(self):
        self._mock_object(self.plugin, '_validate_routes_nexthop')

        router_data = {
            'router': {
                'name': 'test_router2',
                'admin_state_up': True,
                'external_gateway_info': {},
            },
        }

        router = self.plugin.create_router(self.context, router_data)

        routes = [
            {'destination': '10.1.0.0/24', 'nexthop': '192.168.1.1'},
            {'destination': '10.2.0.0/24', 'nexthop': '192.168.1.1'},
            {'destination': '10.3.0.0/24', 'nexthop': '192.168.1.1'}
        ]

        set_routes = []
        update_data = {
            'router': {
                'id': router['id'],
                'routes': set_routes,
            },
        }

        RouteRule = vyatta_utils.RouteRule

        for rules_add in routes:
            rules_add = [rules_add]

            set_routes.extend(rules_add)
            rules_del = set_routes[:-2]
            set_routes[:-2] = []

            self.plugin.update_router(self.context, router['id'], update_data)

            rules_add = tuple(RouteRule(dest_cidr=x['destination'],
                                        next_hop=x['nexthop'])
                              for x in rules_add)
            rules_del = tuple(RouteRule(dest_cidr=x['destination'],
                                        next_hop=x['nexthop'])
                              for x in rules_del)

            self.driver.update_static_routes.assert_called_once_with(
                self.context, router['id'],
                rules_add, rules_del)
            self.driver.reset_mock()

    def test_get_router(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        router = self.plugin.create_router(self.context, router_data)
        router = self.plugin.get_router(self.context, router['id'])
        self.assertTrue(uuidutils.is_uuid_like(router.get('id')))

        self.assertRaises(l3.RouterNotFound, self.plugin.get_router,
                          self.context, uuidutils.generate_uuid())

    def test_delete_router(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        router = self.plugin.create_router(self.context, router_data)
        self.plugin.delete_router(self.context, router['id'])
        self.driver.delete_router.assert_called_once_with(
            self.context, router['id'])
        self.assertRaises(
            l3.RouterNotFound, self.plugin.delete_router,
            self.context, router['id'])

    def test_router_interface_by_subnet(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        router = self.plugin.create_router(self.context, router_data)

        result = self.plugin.add_router_interface(self.context, router['id'], {
            'subnet_id': self.ext_subnet['id'],
        })
        self.driver.attach_interface.assert_called_once_with(
            self.context, router['id'], result['port_id'])

        result = self.plugin.remove_router_interface(
            self.context, router['id'], {
                'subnet_id': self.ext_subnet['id']})

        self.driver.detach_interface.assert_called_once_with(
            self.context, router['id'], result['port_id'])

    def test_router_interface_by_port(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        router = self.plugin.create_router(self.context, router_data)

        self.plugin.add_router_interface(self.context, router['id'], {
            'port_id': self.ext_port['id'],
        })
        self.driver.attach_interface.assert_called_once_with(
            self.context, router['id'], self.ext_port['id'])

        self.plugin.remove_router_interface(
            self.context, router['id'], {
                'port_id': self.ext_port['id']
            })
        self.driver.detach_interface.assert_called_once_with(
            self.context, router['id'], self.ext_port['id'])

    def test_floatingip(self):
        router_data = {
            'router': {'name': 'test_router1', 'admin_state_up': True}}
        router = self.plugin.create_router(self.context, router_data)

        floatingip = self.plugin.create_floatingip(
            self.context,
            {'floatingip': {'floating_network_id': self.ext_net['id']}})
        self.addCleanup(self.plugin.delete_floatingip,
                        self.context, floatingip['id'])
        self.assertTrue(
            floatingip['floating_ip_address'].startswith('10.10.10.'))

        self.plugin.associate_floatingip(
            self.context, router['id'], floatingip)
        self.driver.assign_floating_ip.assert_called_once_with(
            self.context, router['id'], floatingip['floating_ip_address'],
            None)

        self.plugin.disassociate_floatingip(
            self.context, router['id'], floatingip)
        self.driver.unassign_floating_ip.assert_called_once_with(
            self.context, router['id'], floatingip['floating_ip_address'],
            None)

        self.plugin.update_floatingip(self.context, floatingip['id'], {
            'floatingip': {
                'router_id': router['id'],
            }})
        self.driver.assign_floating_ip.assert_called_once_with(
            self.context, router['id'], floatingip['floating_ip_address'],
            None)

CORE_PLUGIN_CLASS = (
    "networking_brocade.vyatta.tests.test_vrouter_neutron_plugin"
    ".TestVRouterNatPlugin")
L3_PLUGIN_CLASS = (
    "networking_brocade.vyatta.vrouter.neutron_plugin.VyattaVRouterMixin")


class TestVRouterNatPlugin(test_l3_plugin.TestL3NatBasePlugin,
                           sg_db_rpc.SecurityGroupServerRpcMixin):
    supported_extension_aliases = ["external-net"]


class VRouterTestCase(test_db_plugin.NeutronDbPluginV2TestCase,
                      test_l3_plugin.L3NatTestCaseBase):
    def setUp(self, core_plugin=None, l3_plugin=None, ext_mgr=None):

        if not core_plugin:
            core_plugin = CORE_PLUGIN_CLASS
        if not l3_plugin:
            l3_plugin = L3_PLUGIN_CLASS

        service_plugins = {'l3_plugin_name': l3_plugin}

        self._mock('eventlet.greenthread.sleep')

        self._mock(
            'networking_brocade.vyatta.vrouter.driver.'
            'VyattaVRouterDriver', FakeVRouterDriver)

        cfg.CONF.set_default('allow_overlapping_ips', True)
        cfg.CONF.set_override('tenant_id', 'tenant_a', 'VROUTER')

        super(VRouterTestCase, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins,
            ext_mgr=test_l3_plugin.L3TestExtensionManager())

        self.setup_notification_driver()

    def _mock(self, target, new=mock.DEFAULT):
        patcher = mock.patch(target, new)
        return patcher.start()

    def test_router_add_interface_ipv6_subnet(self):
        self.skipTest("Fails because router port is created with"
                      " empty device owner")

    def test_router_delete_ipv6_slaac_subnet_inuse_returns_409(self):
        self.skipTest("Fails because router port is created with"
                      " empty device owner")

    def test_router_delete_dhcpv6_stateless_subnet_inuse_returns_409(self):
        self.skipTest("Fails because router port is created with"
                      " empty device owner")

    def test_router_add_gateway_no_subnet(self):
        self.skipTest("Skip because it is not supported.")

    def test_router_specify_id_backend(self):
        self.skipTest("Router id is autogenerated")

    def test_router_update_gateway_upon_subnet_create_max_ips_ipv6(self):
        self.skipTest("Router external gateway supports only one IP address")



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from networking_brocade.vyatta.common import utils


# TODO(asaprykin): Move to parsers
TOKEN_GROUP = 'group'
TOKEN_PARAM = 'param'
TOKEN_END = 'end'


def parse_line(line):
    value = None
    if line.endswith('{'):
        key = line[:-1].strip()
        token = TOKEN_GROUP
    elif line.endswith('}'):
        key = line[:-1].strip()
        token = TOKEN_END
    else:
        token = TOKEN_PARAM
        chunks = line.split(' ', 1)
        key = chunks.pop(0)
        if chunks:
            value = chunks[0]

    return token, key, value


def config_iter(config):
    for line in config.splitlines():
        line = line.strip()
        if line:
            yield line


def parse_group(lines):
    result = utils.MultiDict()

    for line in lines:
        token, key, value = parse_line(line)

        if token == TOKEN_PARAM:
            result.setlistdefault(key).append(value)
        elif token == TOKEN_GROUP:
            result[key] = parse_group(lines)
        else:
            break

    return result


def parse_config(config):
    lines = config_iter(config)
    return parse_group(lines)


class RouterConfig(object):

    EXTERNAL_GATEWAY = 'External_Gateway'
    ROUTER_INTERFACE = 'Router_Interface'

    def __init__(self, data):
        self.data = data

    def interfaces(self):
        try:
            ifaces = self.data['interfaces']
        except KeyError:
            return

        for key, params in six.iteritems(ifaces):
            type_, name = key.split()

            info = utils.MultiDict(params)
            info['name'] = name
            info['type'] = type_

            yield info

    def find_interface(self, name):
        for interface in self.interfaces():
            if interface['name'] == name:
                return interface

    def router_interfaces(self):
        for iface in self.interfaces():
            if iface['descr'] == self.ROUTER_INTERFACE:
                yield iface

    @classmethod
    def from_string(cls, string):
        return cls(parse_config(string))



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import testtools

from networking_brocade.vyatta.common import utils


class TestMultiDict(testtools.TestCase):

    def setUp(self):
        super(TestMultiDict, self).setUp()
        # Create from list of tuples
        mapping = [('a', 1), ('b', 2), ('a', 2), ('d', 3),
                   ('a', 1), ('a', 3), ('d', 4), ('c', 3)]
        self.md = utils.MultiDict(mapping)

    def test_init(self):
        md = utils.MultiDict()
        self.assertIsInstance(md, collections.MutableMapping)

        # Create from dict
        mapping = {'a': 1, 'b': 2, 'c': 3}
        md = utils.MultiDict(mapping)
        self.assertEqual(md['a'], 1)
        self.assertEqual(md.getlist('a'), [1])

    def test_getitem(self):
        # __getitem__
        self.assertEqual(self.md['a'], 1)
        self.assertEqual(self.md['c'], 3)
        with testtools.ExpectedException(KeyError):
            self.md['e']

        # get
        self.assertEqual(self.md.get('a'), 1)
        self.assertEqual(self.md.get('e'), None)

        # getlist
        self.assertEqual(self.md.getlist('a'), [1, 2, 1, 3])
        self.assertEqual(self.md.getlist('d'), [3, 4])
        self.assertEqual(self.md.getlist('x'), [])

    def test_setitem(self):
        # __setitem__
        self.md['a'] = 42
        self.assertEqual(self.md['a'], 42)
        self.assertEqual(self.md.getlist('a'), [42])

        # setlist
        self.md.setlist('a', [1, 2, 3])
        self.assertEqual(self.md['a'], 1)
        self.assertEqual(self.md.getlist('a'), [1, 2, 3])

        # check that setlist does not affects initial list
        lst = [1, 2, 3]
        self.md.setlist('a', lst)
        self.md.add('a', 42)
        self.assertEqual(lst, [1, 2, 3])

        # setdefault
        # TODO(asaprykin): Check setdefault without arg
        self.assertEqual(self.md.setdefault('u', 32), 32)
        self.assertEqual(self.md.getlist('u'), [32])

    def test_delitem(self):
        # __delitem__
        del self.md['a']
        self.assertNotIn('a', self.md)
        with testtools.ExpectedException(KeyError):
            del self.md['x']

    def test_setlistdefault(self):

        self.assertEqual(self.md.setlistdefault('a'), [1, 2, 1, 3])
        self.assertEqual(self.md.setlistdefault('d'), [3, 4])

        self.assertEqual(self.md.setlistdefault('u1'), [])
        self.assertEqual(self.md.setlistdefault('u3', [42]), [42])

        with testtools.ExpectedException(TypeError):
            self.md.setlistdefault('u2', False)
        with testtools.ExpectedException(TypeError):
            self.md.setlistdefault('u2', 32)



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
    Connector class used to connect to device.
    Decides which connector to use - TELNET or SSH, based on
    the argument passed
"""
import networking_brocade.mlx.ml2.commands as Commands
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class DeviceConnector(object):

    """
    Connector class used to connect to device.
    Decides which connector to use - TELNET or SSH, based on
    the argument passed
    """

    def __init__(self, device_info):
        self.host = device_info.get('address')
        self.username = device_info.get('username')
        self.password = device_info.get('password')
        self.transport = device_info.get('transport')

    def enter_configuration_mode(self):
        """
        Enter configuration mode. First it enters enable mode
        and then to configuration mode. There should be no
        Enable password.
        """

        self.write(Commands.ENABLE_TERMINAL_CMD)
        self.write(Commands.CONFIGURE_TERMINAL_CMD)

    def exit_configuration_mode(self):
        """
        Exit Configuration mode.
        """
        self.send_exit(2)

    def exit_from_device(self):
        """
        Exit Configuration mode and device.
        """
        self.exit_configuration_mode()
        self.send_exit(1)

    def create_vlan(self, vlanid, ports):
        """
        Creates VLAN and tags the ports to the created VLAN
        :param:vlanid: vlan id to be created
        :param:ports: ports to be tagged to the created vlan
        :returns: Response from the device as a list
        """
        LOG.debug("DeviceConnector:create_vlan:Creating vlan %(vlanid)s on "
                  "device %(host)s", {'vlanid': vlanid, 'host': self.host})
        self.enter_configuration_mode()
        self.write(
            Commands.CONFIGURE_VLAN.format(
                vlan_id=vlanid))
        LOG.debug(
            "Created VLAN with id %(vlanid)s on device %(host)s", {
                'vlanid': vlanid, 'host': self.host})
        for port in ports:
            self.tag_port(port)
            LOG.debug("tagged port %(port)s", {'port': port})
        self.send_exit(1)
        self.exit_from_device()
        return self.read_response()

    def tag_port(self, port):
        """Tag Ports to the vlan

        :param:port: Port to be tagged to the vlan
        """
        LOG.debug("DeviceConnector:tag_port:Tagging port %(portid)s on device"
                  " %(host)s", {'portid': port, 'host': self.host})
        self.write(
            Commands.CONFIGURE_ETHERNET.format(
                port_number=port))

    def delete_vlan(self, vlan_id):
        """Deletes the VLAN from the device

        :param:vlan_id: vlan id to be deleted
        :returns: Response from the device as a list
        """
        LOG.debug("DeviceConnector:delete_vlan:Deleting VLAN with id "
                  "%(vlan_id)s on device %(host)s", {'vlan_id': vlan_id,
                                                     'host': self.host})
        self.enter_configuration_mode()
        self.write(
            Commands.DELETE_CONFIGURED_VLAN.format(
                vlan_id=vlan_id))
        LOG.debug(
            "Deleted VLAN with id %(vlan_id)s on device %(host)s", {
                'vlan_id': vlan_id, 'host': self.host})
        self.exit_from_device()
        return self.read_response()

    def get_version(self):
        """Get Firmware Version

        :returns: Response from the device as a string
        """
        LOG.debug("DeviceConnector:get_version:Executing show version for "
                  "device %(host)s", {'host': self.host})
        self.write(Commands.SHOW_VERSION)
        self.write(Commands.CTRL_C)
        self.send_exit(1)
        return self.read_response(read_lines=False)

    def create_l3_router(self, vlan_id, gateway_ip_cidr):
        """Configures a Router Interface interface

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        :param:gateway_ip_cidr: Gateway IP address with network length.
        :returns: Response from the device as a list
        """

        self.enter_configuration_mode()
        LOG.debug(("DeviceConnector:create_l3_router:Configuring router "
                   "interface with id %(vlanid)s on device %(host)s"),
                  {'vlanid': vlan_id, 'host': self.host})
        self._create_router_interface(vlan_id)
        self._configure_ipaddress(vlan_id, gateway_ip_cidr)
        LOG.debug(("DeviceConnector:create_l3_router:Configured router "
                   "interface with id %(vlanid)s on device %(host)s"),
                  {'vlanid': vlan_id, 'host': self.host})
        self.exit_from_device()
        return self.read_response()

    def _create_router_interface(self, vlan_id):
        """Create VLAN and Router Interface

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        """
        LOG.debug("DeviceConnector:_create_router_interface:Creating l3 "
                  "router interface %(vlanid)s on device "
                  "%(host)s", {'vlanid': vlan_id,
                               'host': self.host})
        self.write(
            Commands.CONFIGURE_VLAN.format(
                vlan_id=vlan_id))
        self.write(
            Commands.CONFIGURE_ROUTER_INTERFACE.format(
                vlan_id=vlan_id))
        LOG.debug("DeviceConnector:_create_router_interface:Created l3 "
                  "router interface %(vlanid)s on device "
                  "%(host)s", {'vlanid': vlan_id,
                               'host': self.host})
        self.send_exit(1)

    def _configure_ipaddress(self, vlan_id, gateway_ip_cidr):
        """Assigns Gateway ip for the configured vlan

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        :param:gateway_ip_cidr: Gateway IP address with network length.
        """
        LOG.debug("DeviceConnector:_configure_ipaddress:Assigning IP address"
                  " %(ipaddr)s to the router interface %(vlanid)s on device"
                  " %(host)s", {'ipaddr': gateway_ip_cidr,
                                'vlanid': vlan_id,
                                'host': self.host})
        self.write(
            Commands.CONFIGURE_INTERFACE.format(
                vlan_id=vlan_id))
        self.write(
            Commands.CONFIGURE_GATEWAY_IP.format(
                gateway_ip_addr=gateway_ip_cidr))
        LOG.debug("DeviceConnector:_configure_ipaddress:Assigned IP address"
                  " %(ipaddr)s to the router interface %(vlanid)s on device"
                  " %(host)s", {'ipaddr': gateway_ip_cidr,
                                'vlanid': vlan_id,
                                'host': self.host})
        self.send_exit(1)

    def delete_l3_router(self, vlan_id):
        """
        Deletes Router Interface with the vlan id

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        :returns: Response from the device as a list
        """

        self.enter_configuration_mode()

        LOG.debug(("DeviceConnector:delete_l3_router:Deleting router interface"
                   " %(vlan_id)s from vlan %(vlan_id)s on device %(host)s"),
                  {'vlan_id': vlan_id, 'host': self.host})
        self.write(
            Commands.CONFIGURE_VLAN.format(vlan_id=vlan_id))
        self.write(
            Commands.DELETE_ROUTER_INTERFACE.format(
                vlan_id=vlan_id))
        self.send_exit(1)

        LOG.debug(("DeviceConnector:delete_l3_router:Deleted router interface"
                   " %(vlan_id)s from vlan %(vlan_id)s on device %(host)s"),
                  {'vlan_id': vlan_id, 'host': self.host})

        self.exit_from_device()
        return self.read_response()



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

import mock
from neutron.db import api as db
from neutron.i18n import _LI
from neutron.tests.unit.extensions import test_l3
from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
L3_SVC_PLUGIN = ('neutron.services.l3_router.'
                 'brocade.l3_router_plugin.BrocadeSVIPlugin')


class BrocadeSVIPlugin_TestCases(test_l3.TestL3NatBasePlugin):

    def setUp(self):

        def mocked_brocade_init(self):
            LOG.debug("brocadeSVIPlugin::mocked_brocade_init()")

            self._switch = {'address': cfg.CONF.ml2_brocade.address,
                            'username': cfg.CONF.ml2_brocade.username,
                            'password': cfg.CONF.ml2_brocade.password,
                            'rbridge_id': cfg.CONF.ml2_brocade.rbridge_id
                            }
            LOG.info(_LI("rbridge id %s"), self._switch['rbridge_id'])
            self._driver = mock.MagicMock()

        self.l3_plugin = importutils.import_object(L3_SVC_PLUGIN)
        with mock.patch.object(self.l3_plugin,
                               'brocade_init', new=mocked_brocade_init):
            super(BrocadeSVIPlugin_TestCases, self).setUp()
        self.context = oslo_context.get_admin_context()
        self.context.session = db.get_session()


class TestBrocadeSVINatBase(test_l3.L3NatExtensionTestCase,
                            BrocadeSVIPlugin_TestCases):
    pass



# Copyright 2015 Brocade Communications Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Implementation of Brocade ML2 Mechanism driver for ICX and MLX."""

from neutron.i18n import _LE
from neutron.i18n import _LI
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api
from oslo_log import log as logging
from oslo_utils import importutils

MECHANISM_VERSION = 0.1
BROCADE_CONFIG = ("networking_brocade.mlx.ml2.fi_ni.brcd_config."
                  "ML2BrocadeConfig")
DRIVER_FACTORY = ("networking_brocade.mlx.ml2.fi_ni."
                  "driver_factory.BrocadeDriverFactory")
LOG = logging.getLogger(__name__)


class BrocadeFiNiMechanism(driver_api.MechanismDriver):

    """
    ML2 Mechanism driver for Brocade ICX and MLX devices. This is the upper
    layer driver class that interfaces to lower layer (SSH/TELNET) below.

    """

    def __init__(self):
        self._driver = None
        self._physical_networks = None
        self._switch = None
        self._driver_map = {}
        self._devices = {}
        self.initialize()

    def initialize(self):
        """Initialize of variables needed by this class."""

        self._devices, self._physical_networks = importutils.import_object(
            BROCADE_CONFIG).create_brocade_dictionary()

    def create_network_precommit(self, mech_context):
        """No-op."""
        pass

    def create_network_postcommit(self, mech_context):
        """Create VLAN on the switch.

        :param:mech_context: Details about the network to be created
        :raise: Exception
        """

        LOG.debug("create_network_postcommit: called")
        network = mech_context.current
        network_id = network['id']
        tenant_id = network['tenant_id']
        segments = mech_context.network_segments
        network_type = segments[0]['network_type']
        vlan_id = segments[0]['segmentation_id']
        physical_network = segments[0]['physical_network']
        if physical_network not in self._physical_networks:
            LOG.exception(_LE("BrocadeFiNiMechanism: Failed to create network."
                              " Network cannot be created in the configured "
                              "physical network %(physnet)s"),
                          {'physnet': physical_network})
            raise ml2_exc.MechanismDriverError(method='create_network_postcomm'
                                               'it')
        if network_type != 'vlan':
            LOG.exception(_LE("BrocadeFiNiMechanism: Failed to create network "
                              "for network type %(nw_type)s. Only network type"
                              " vlan is supported"), {'nw_type': network_type})
            raise ml2_exc.MechanismDriverError(method='create_network_postcomm'
                                               'it')
        try:
            devices = self._physical_networks.get(physical_network)
            for device in devices:
                device_info = self._devices.get(device)
                address = device_info.get('address')
                driver = None
                try:
                    driver = self._get_driver(device)
                except Exception as e:
                    LOG.exception(_LE("BrocadeFiNiMechanism: create_network"
                                      "_postcommit failed while configuring "
                                      "device %(host)s exception=%(error)s"),
                                  {'host': address,
                                   'error': e.args})
                    raise ml2_exc.MechanismDriverError(method='create_network_'
                                                       'postcommit')
                # Proceed only if the driver is not None
                if driver is not None:
                    driver.create_network(
                        vlan_id,
                        device_info.get("ports").split(","))
        except Exception as e:
            LOG.exception(
                _LE("Brocade FI/NI driver: create_network_postcommit failed"
                    "Error = %(error)s"), {'error': e.args})
            raise ml2_exc.MechanismDriverError(method='create_network_postcomm'
                                               'it')
        LOG.info(_LI("BrocadeFiNiMechanism:created_network_postcommit: "
                     "%(network_id)s of network type = %(network_type)s with "
                     "vlan = %(vlan_id)s for tenant %(tenant_id)s"),
                 {'network_id': network_id,
                  'network_type': network_type,
                  'vlan_id': vlan_id,
                  'tenant_id': tenant_id})

    def delete_network_precommit(self, mech_context):
        """No-op."""
        pass

    def delete_network_postcommit(self, mech_context):
        """Delete network which translates to removing vlan
        from the switch.

        :param:mech_context: Details about the network to be created
        :raise: MechanismDriverError
        """

        LOG.debug("delete_network_postcommit: called")
        network = mech_context.current
        network_id = network['id']
        vlan_id = network['provider:segmentation_id']
        tenant_id = network['tenant_id']
        segments = mech_context.network_segments
        segment = segments[0]
        network_type = segment['network_type']
        physical_network = segment['physical_network']
        if physical_network not in self._physical_networks:
            LOG.exception(_LE("BrocadeFiNiMechanism: Failed to delete network."
                              " Network cannot be deleted in the configured "
                              "physical network %(physnet)s"),
                          {'physnet': physical_network})
            raise ml2_exc.MechanismDriverError(method='delete_network_postcomm'
                                               'it')
        if network_type != 'vlan':
            LOG.exception(_LE("BrocadeFiNiMechanism: Failed to delete network "
                              "for network type %(nw_type)s. Only network type"
                              " vlan is supported"), {'nw_type': network_type})
            raise ml2_exc.MechanismDriverError(method='delete_network_postcomm'
                                               'it')
        try:
            devices = self._physical_networks.get(physical_network)
            for device in devices:
                driver = self._get_driver(device)
                driver.delete_network(vlan_id)
        except Exception:
            LOG.exception(
                _LE("BrocadeFiNiMechanism: failed to delete network"))
            raise ml2_exc.MechanismDriverError(method='delete_network_postcomm'
                                               'it')

        LOG.info(_LI("BrocadeFiNiMechanism: delete network (postcommit): "
                     "%(network_id)s with vlan = %(vlan_id)s for tenant "
                     "%(tenant_id)s"), {'network_id': network_id,
                                        'vlan_id': vlan_id,
                                        'tenant_id': tenant_id})

    def update_network_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def update_network_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def create_port_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def create_port_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def delete_port_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def delete_port_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def update_port_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_port_precommit(self: called")

    def update_port_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_port_postcommit: called")

    def create_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("create_subnetwork_precommit: called")

    def create_subnet_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("create_subnetwork_postcommit: called")

    def delete_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("delete_subnetwork_precommit: called")

    def delete_subnet_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("delete_subnetwork_postcommit: called")

    def update_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_subnet_precommit(self: called")

    def update_subnet_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_subnet_postcommit: called")

    def _get_driver(self, device):
        """
        Gets the driver based on the firmware version of the device

        :param:device: A dictionary which contains details of all the devices
            parsed from the configuration template
        :raise: Exception
        """
        driver = self._driver_map.get(device)
        if driver is None:
            driver_factory = importutils.import_object(DRIVER_FACTORY)
            device_info = self._devices.get(device)
            address = device_info.get('address')
            try:
                driver = driver_factory.get_driver(device_info)
            except Exception as e:
                LOG.exception(_LE("BrocadeFiNiMechanism:_get_driver failed for"
                                  "device %(host)s: Error = %(error)s"),
                              {'host': address,
                               'error': e.args})
                raise Exception(_("BrocadeFiNiMechanism: "
                                  "Failed to get driver"))
            self._driver_map.update({device: driver})
        return driver



# Copyright 2014 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Implentation of Brocade SVI service Plugin."""


from networking_brocade.vdx.ml2driver.nos import nosdriver as driver
from neutron.common import constants as l3_constants
from neutron.i18n import _LE
from neutron.i18n import _LI
from neutron.plugins.ml2 import db
from neutron.plugins.ml2.drivers.brocade.db import models as brocade_db
from neutron.services.l3_router import l3_router_plugin as router
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils


DEVICE_OWNER_ROUTER_INTF = l3_constants.DEVICE_OWNER_ROUTER_INTF
DEVICE_OWNER_ROUTER_GW = l3_constants.DEVICE_OWNER_ROUTER_GW
DEVICE_OWNER_FLOATINGIP = l3_constants.DEVICE_OWNER_FLOATINGIP

ML2_BROCADE = [cfg.StrOpt('address', default='',
                          help=_('The address of the host to SSH to')),
               cfg.StrOpt('username', default='admin',
                          help=_('The SSH username to use')),
               cfg.StrOpt('password', default='password', secret=True,
                          help=_('The SSH password to use')),
               cfg.StrOpt('rbridge_id', default=1,
                          help=_('Rbridge id of provider edge router(s)')),
               ]

cfg.CONF.register_opts(ML2_BROCADE, "ml2_brocade")

LOG = logging.getLogger(__name__)


class BrocadeSVIPlugin(router.L3RouterPlugin):
    """Brocade SVI service Plugin."""

    def __init__(self):
        """Initialize Brocade Plugin

        Specify switch address and db configuration.
        """
        super(BrocadeSVIPlugin, self).__init__()
        self._switch = None
        self._driver = None
        self.brocade_init()

    def brocade_init(self):
        """Brocade specific initialization."""
        LOG.debug("brocadeSVIPlugin::brocade_init()")

        self._switch = {'address': cfg.CONF.ml2_brocade.address,
                        'username': cfg.CONF.ml2_brocade.username,
                        'password': cfg.CONF.ml2_brocade.password,
                        'rbridge_id': cfg.CONF.ml2_brocade.rbridge_id
                        }
        self._driver = driver.NOSdriver()
        LOG.info(_LI("rbridge id %s"), self._switch['rbridge_id'])

    def create_router(self, context, router):
        """Creates a vrf on NOS device."""
        LOG.debug("BrocadeSVIPlugin.create_router called: ")
        with context.session.begin(subtransactions=True):
            new_router = super(BrocadeSVIPlugin, self).create_router(context,
                                                                     router)
            # Router on VDX
            try:
                switch = self._switch
                self._driver.create_router(switch['address'],
                                           switch['username'],
                                           switch['password'],
                                           switch['rbridge_id'],
                                           str(new_router['id']))
            except Exception:
                with excutils.save_and_reraise_exception():
                    with context.session.begin(subtransactions=True):
                        super(BrocadeSVIPlugin, self).delete_router(
                            context,
                            new_router['id'])

        LOG.debug("BrocadeSVIPlugin.create_router: "
                  "router created on VDX switch")
        return new_router

    def delete_router(self, context, router_id):
        """Delete a vrf on NOS device."""
        router = super(BrocadeSVIPlugin, self).get_router(context, router_id)
        super(BrocadeSVIPlugin, self).delete_router(context, router_id)

        switch = self._switch
        self._driver.delete_router(switch['address'],
                                   switch['username'],
                                   switch['password'],
                                   switch['rbridge_id'],
                                   str(router['id']))

    def add_router_interface(self, context, router_id, interface_info):
        """creates svi on NOS device and assigns ip addres to SVI."""
        LOG.debug("BrocadeSVIPlugin.add_router_interface on VDX: "
                  "router_id=%(router_id)s "
                  "interface_info=%(interface_info)r",
                  {'router_id': router_id, 'interface_info': interface_info})

        with context.session.begin(subtransactions=True):

            info = super(BrocadeSVIPlugin, self).add_router_interface(
                context, router_id, interface_info)

            port = db.get_port(context.session, info["port_id"])

            # shutting down neutron port to allow NOS to do Arp/Routing
            port['admin_state_up'] = False
            port['port'] = port
            self._core_plugin.update_port(context, info["port_id"], port)

            interface_info = info
            subnet = self._core_plugin._get_subnet(context,
                                                   interface_info["subnet_id"])
            cidr = subnet["cidr"]
            net_addr, net_len = self.net_addr(cidr)
            gateway_ip = subnet["gateway_ip"]
            network_id = subnet['network_id']
            bnet = brocade_db.get_network(context, network_id)
            vlan_id = bnet['vlan']
            gateway_ip_cidr = gateway_ip + '/' + str(net_len)
            LOG.debug("Allocated cidr %(cidr)s from the pool, "
                      "network_id %(net_id)s "
                      "bnet %(bnet)s "
                      "vlan %(vlan_id)d ", {'cidr': gateway_ip_cidr,
                                            'net_id': network_id,
                                            'bnet': bnet,
                                            'vlan_id': int(vlan_id)})
            port_filters = {'network_id': [network_id],
                            'device_owner': [DEVICE_OWNER_ROUTER_INTF]}
            port_count = self._core_plugin.get_ports_count(context,
                                                           port_filters)
            LOG.info(_LI("BrocadeSVIPlugin.add_router_interface ports_count "
                         "%d"),
                     port_count)

            # port count is checked against 2 since the current port is already
            # added to db
            if port_count == 2:
                # This subnet is already part of some router
                # (this is not supported in this version of brocade svi plugin)
                msg = _("BrocadeSVIPlugin: adding redundant router interface "
                        "is not supported")
                LOG.error(msg)
                raise Exception(msg)

        try:
            switch = self._switch
            self._driver.create_svi(switch['address'],
                                    switch['username'],
                                    switch['password'],
                                    switch['rbridge_id'],
                                    vlan_id,
                                    gateway_ip_cidr,
                                    str(router_id))
        except Exception:
            LOG.error(_LE("Failed to create Brocade resources to add router "
                          "interface. info=%(info)s, router_id=%(router_id)s"),
                      {"info": info, "router_id": router_id})
            with excutils.save_and_reraise_exception():
                with context.session.begin(subtransactions=True):
                    self.remove_router_interface(context, router_id,
                                                 interface_info)
        return info

    def remove_router_interface(self, context, router_id, interface_info):
        """Deletes svi from NOS device."""
        LOG.debug("BrocadeSVIPlugin.remove_router_interface called: "
                  "router_id=%(router_id)s "
                  "interface_info=%(interface_info)r",
                  {'router_id': router_id, 'interface_info': interface_info})

        with context.session.begin(subtransactions=True):
            info = super(BrocadeSVIPlugin, self).remove_router_interface(
                context, router_id, interface_info)
            try:
                subnet = self._core_plugin._get_subnet(context,
                                                       info['subnet_id'])
                cidr = subnet['cidr']
                net_addr, net_len = self.net_addr(cidr)
                gateway_ip = subnet['gateway_ip']
                network_id = subnet['network_id']
                bnet = brocade_db.get_network(context, network_id)
                vlan_id = bnet['vlan']
                gateway_ip_cidr = gateway_ip + '/' + str(net_len)
                LOG.debug("remove_router_interface removed cidr %(cidr)s"
                          " from the pool,"
                          " network_id %(net_id)s bnet %(bnet)s"
                          " vlan %(vlan_id)d",
                          {'cidr': gateway_ip_cidr,
                           'net_id': network_id,
                           'bnet': bnet,
                           'vlan_id': int(vlan_id)})
                switch = self._switch
                self._driver.delete_svi(switch['address'],
                                        switch['username'],
                                        switch['password'],
                                        switch['rbridge_id'],
                                        vlan_id,
                                        gateway_ip_cidr,
                                        str(router_id))
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Fail remove of interface from brocade "
                                  "router interface. info=%(info)s, "
                                  "router_id=%(router_id)s"),
                              {"info": info, "router_id": router_id})
        return True

    @staticmethod
    def net_addr(addr):
        """Get network address prefix and length from a given address."""
        if addr is None:
            return None, None
        nw_addr, nw_len = addr.split('/')
        nw_len = int(nw_len)
        return nw_addr, nw_len



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import functools
import itertools
import re

from networking_brocade.vyatta.common import exceptions as v_exc

HWADDR = r'(?:[a-zA-Z0-9]{2}:){5}[a-zA-Z0-9]{2}'
IPv4_ADDR = r'(?:\d+\.){3}\d+'


EMPTY_LINE = re.compile(r'^\s+$')
IFACE = re.compile(r'^(\w+):')
LINK_ETHER = re.compile(r'\s+link/ether\s+({0})'.format(HWADDR))
IP_ADDR = re.compile(r'\s+inet\s+({0})'.format(IPv4_ADDR))


def parse_interfaces(output):
    ifaces = []
    info = {}
    for line in output.splitlines():
        if not line or re.match(EMPTY_LINE, line):
            continue

        m = re.match(IFACE, line)
        if m:
            if info:
                ifaces.append(info)
            name = m.group(1)
            info = dict(name=name, ip_addrs=[], mac_address=None)
            continue

        m = re.match(LINK_ETHER, line)
        if m:
            info['mac_address'] = m.group(1).lower()

        m = re.match(IP_ADDR, line)
        if m:
            info['ip_addrs'].append(m.group(1))

    if info:
        ifaces.append(info)

    return ifaces


class TableParser(collections.Iterable, collections.Sized):
    idx_generator = itertools.count()
    _S_EMPTY = next(idx_generator)
    _S_TITLE_SEPARATOR_INTRO = next(idx_generator)
    _S_TITLE_SEPARATOR_FIELD = next(idx_generator)
    _S_TITLE_SEPARATOR_SPACE = next(idx_generator)
    _S_EXTRACT_HEADERS = next(idx_generator)
    _S_VALUE_LINE = next(idx_generator)
    _S_STORE_RESULT = next(idx_generator)
    del idx_generator

    def __init__(self):
        collections.Iterable.__init__(self)
        self.results = list()

    def __iter__(self):
        return iter(self.results)

    def __len__(self):
        return len(self.results)

    def __call__(self, stream):
        stream = TableTokenizer(stream)

        result = None
        lines_buffer = list()
        fields = list()
        offs = 0

        state = self._S_EMPTY

        for chunk in stream:
            if state == self._S_EMPTY:
                stream.revert(chunk)
                chunk = self._read_until_eol(stream)
                chunk = ''.join(chunk)
                if not chunk or chunk.isspace():
                    lines_buffer = list()
                    continue

                offs = 0
                fields = list()
                lines_buffer.append(chunk)

                state = self._S_TITLE_SEPARATOR_INTRO
                continue

            elif state == self._S_TITLE_SEPARATOR_INTRO:
                if chunk == '\n':
                    lines_buffer = list()
                    state = self._S_EMPTY
                    continue

                if chunk.isspace():
                    offs += len(chunk)
                    continue

                stream.revert(chunk)
                state = self._S_TITLE_SEPARATOR_FIELD
                continue

            elif state == self._S_TITLE_SEPARATOR_FIELD:
                if chunk == '\n':
                    stream.revert(None)
                    state = self._S_EXTRACT_HEADERS
                    continue

                if set(chunk) != set('-'):
                    raise ValueError((
                        'Unexpected chunk: {0}. Expect sequence of \'-\' '
                        'characters').format(chunk))

                fields.append((offs, len(chunk)))
                offs += len(chunk)
                state = self._S_TITLE_SEPARATOR_SPACE
                continue

            elif state == self._S_TITLE_SEPARATOR_SPACE:
                if chunk == '\n':
                    stream.revert(None)
                    state = self._S_EXTRACT_HEADERS
                    continue

                if not chunk.isspace():
                    raise ValueError((
                        'Unexpected chunk: {0}. Expect '
                        'whitespace.').format(chunk))

                offs += len(chunk)
                state = self._S_TITLE_SEPARATOR_FIELD
                continue

            elif state == self._S_EXTRACT_HEADERS:
                if not fields:
                    lines_buffer = list()
                    state = self._S_EMPTY
                    continue

                result = self._extract_field_titles(lines_buffer, fields)
                lines_buffer = list()
                state = self._S_VALUE_LINE
                continue

            elif state == self._S_VALUE_LINE:
                stream.revert(chunk)
                chunk = self._read_until_eol(stream)
                chunk = ''.join(chunk)
                if not chunk:
                    stream.revert(None)
                    state = self._S_STORE_RESULT
                    continue

                self._extract_field_values([chunk], fields, result)
                state = self._S_VALUE_LINE
                continue

            elif state == self._S_STORE_RESULT:
                self.results.append(result)

                result = None
                lines_buffer = list()
                state = self._S_EMPTY
                continue

            else:
                raise RuntimeError('Reach unreachable point.')

    def _extract_field_titles(self, buffer, ranges):
        return _Table(self._extract_data_by_ranges(buffer, ranges))

    def _extract_field_values(self, buffer, ranges, table):
        values = self._extract_data_by_ranges(buffer, ranges)
        table.add(values)

    def _read_until_eol(self, stream):
        result = list()
        try:
            while True:
                chunk = next(stream)
                if chunk == '\n':
                    break
                result.append(chunk)
        except StopIteration:
            pass
        return result

    def _extract_data_by_ranges(self, buffer, ranges):
        result = list()
        for offs, length in ranges:
            offs_end = offs + length

            if offs < 0 or offs_end <= offs:
                raise ValueError(
                    'Invalid extract range offs {0} end {1}'.format(
                        offs, offs_end))

            value = list()
            for line in buffer:
                if len(line) < offs:
                    raise ValueError(
                        'Unable to extract value from {0}, offs {1} end {2}: '
                        'payload too short'.format(line, offs, offs_end))
                # Ignore offs_end, instead use delimiter to extract value
                # Fix for mangled ip-addr values from 'sh vpn ipsec sa'
                # parsing
                value.append(line[offs:].split()[0])
            value = ''.join(value)
            value = value.strip()
            result.append(value)

        return tuple(result)


class _ReversibleIterator(collections.Iterator):
    def __init__(self):
        collections.Iterator.__init__(self)
        self._reverse_storage = list()

    def next(self):
        if self._reverse_storage:
            return self._reverse_storage.pop(0)
        return self._next()

    @abc.abstractmethod
    def _next(self):
        raise StopIteration

    def revert(self, *items):
        self._reverse_storage[:0] = reversed(items)


class TableTokenizer(_ReversibleIterator):
    def __init__(self, stream):
        _ReversibleIterator.__init__(self)
        self.character_stream = _SourceWrapper(stream)

    def _next(self):
        chunk = list()
        char = next(self.character_stream)
        if char == '\n':
            return char

        value_type = type(char)
        if char.isspace():
            end_condition = lambda ch: ch != '\n' and ch.isspace()
        else:
            end_condition = lambda ch: not ch.isspace()

        while end_condition(char):
            chunk.append(char)
            try:
                char = next(self.character_stream)
            except StopIteration:
                break
        else:
            self.character_stream.revert(char)

        return value_type('').join(chunk)


class _SourceWrapper(_ReversibleIterator):
    _chunk = iter(tuple())

    def __init__(self, stream):
        _ReversibleIterator.__init__(self)
        self._stream = iter(stream)

    def _next(self):
        while True:
            try:
                return next(self._chunk)
            except StopIteration:
                chunk = next(self._stream)
                self._chunk = iter(chunk)


class _Table(collections.Iterable):
    def __init__(self, headers):
        collections.Iterable.__init__(self)
        self.headers = tuple(headers)
        self.rows = list()

    def __iter__(self):
        row_factory = functools.partial(_TableRow, self.headers)
        return itertools.imap(row_factory, self.rows)

    def __len__(self):
        return len(self.rows)

    def add(self, payload):
        payload = tuple(payload)
        assert len(payload) == len(self.headers)
        self.rows.append(payload)


class _TableRow(collections.Iterable):
    def __init__(self, headers, payload):
        collections.Iterable.__init__(self)
        if len(headers) != len(payload):
            raise ValueError(
                'Invalid table row: fields number is not equal headers '
                'number.')
        self.headers = headers
        self.payload = payload

    def __iter__(self):
        return itertools.izip(self.headers, self.payload)

    def cell_by_idx(self, idx):
        try:
            value = self.payload[idx]
        except IndexError:
            raise v_exc.TableCellNotFound
        return value

    def cell_by_name(self, name):
        idx = self.headers.index(name)
        return self.cell_by_idx(idx)

    def get_column(self, idx_or_name):
        try:
            if not isinstance(idx_or_name, (int, long)):
                raise ValueError
            value = self.cell_by_idx(idx_or_name)
        except (ValueError, v_exc.TableCellNotFound):
            value = self.cell_by_name(idx_or_name)
        return value



# Copyright 2015 Brocade Communications Systems, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
from networking_brocade.mlx.services.l3_router.brocade import (
    l3_router_plugin as brocadel3routerplugin)
from neutron import context
from neutron.db import l3_db
from neutron.i18n import _LE
from neutron.services.l3_router import l3_router_plugin as router
from neutron.tests import base
from oslo_utils import importutils


MECHANISM_NAME = ('networking_brocade.mlx.services.l3_router.brocade.'
                  'l3_router_plugin.BrocadeRouterPlugin')
VE = 've '
config_map = {}
vlan_map = {}
interface_map = {}

device_map = {
    'mlx': {
        'address': '2.2.2.2',
        'username': 'admin',
        'password': 'pass',
        'physical_networks': 'physnet1',
        'ports': '2/1,2/2',
        'os_type': 'NI'
    }
}
add_config_map = {
    '2.2.2.2': {
        '402': {'ve 402': '12.0.0.1/24'
                }
    }
}

delete_config_map = {
    '2.2.2.2': {
        '402': {}
    }
}

interface_info = {
    'subnet_id': 'subnet1',
    'port_id': 'port1'}

vlan_id = '402'
gateway_ip_cidr = '12.0.0.1/24'
ROUTER = 'router1'
SUBNET = 'subnet1'
PORT = 'port1'

INVALID_INPUT = "Invalid input"
RANGE_ERROR = "outside of allowed max"


class TestBrocadeL3RouterPlugin(base.BaseTestCase, router.L3RouterPlugin):

    """
    Test Brocade L3 Router FI/NI plugin.
    """

    def setUp(self):
        _mechanism_name = MECHANISM_NAME

        def mocked_initialize(self):
            self._devices = device_map

        with mock.patch.object(brocadel3routerplugin
                               .BrocadeRouterPlugin,
                               'brocade_init', new=mocked_initialize):
            super(TestBrocadeL3RouterPlugin, self).setUp()
            self.driver = importutils.import_object(_mechanism_name)

    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_driver')
    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_network_info')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_core_plugin')
    @mock.patch.object(router.L3RouterPlugin, 'add_router_interface')
    def test_add_router_interface(
        self, mock_super, mock_core_plugin, mock_network_info,
            mock_driver):

        mock_driver.side_effect = self.side_effect
        mech_ctx = self._get_network_context('physnet1', 'vlan')
        ctx = mech_ctx._plugin_context
        ctx.session.begin = mock.MagicMock()
        mock_super.returnValue = interface_info
        mock_core_plugin.side_effect = mock.MagicMock()
        mock_network_info.side_effect = self.side_effect_network_info
        self.driver.add_router_interface(ctx, ROUTER,
                                         interface_info)
        self.assertDictSupersetOf(config_map, add_config_map)

    @mock.patch.object(router.L3RouterPlugin, 'remove_router_interface')
    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_driver')
    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_network_info')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_core_plugin')
    @mock.patch.object(router.L3RouterPlugin, 'add_router_interface')
    def test_add_router_interface_exception(
        self, mock_super, mock_core_plugin, mock_network_info,
            mock_driver, mock_super_remove):

        mock_driver.side_effect = self.side_effect_error
        mech_ctx = self._get_network_context('physnet1', 'vlan')
        ctx = mech_ctx._plugin_context
        ctx.session.begin = mock.MagicMock()
        mock_super.returnValue = interface_info
        mock_core_plugin.side_effect = mock.MagicMock()
        mock_network_info.side_effect = self.side_effect_network_info
        mock_super_remove.returnValue = interface_info
        self.assertRaisesRegexp(
            Exception, (_LE("BrocadeRouterPlugin")),
            self.driver.add_router_interface,
            ctx, ROUTER, interface_info)

    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_driver')
    @mock.patch.object(brocadel3routerplugin.BrocadeRouterPlugin,
                       '_get_network_info')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_core_plugin')
    @mock.patch.object(router.L3RouterPlugin, 'remove_router_interface')
    def test_remove_router_interface(
        self, mock_super, mock_core_plugin, mock_network_info,
            mock_driver):

        mock_driver.side_effect = self.side_effect
        mech_ctx = self._get_network_context('physnet1', 'vlan')
        ctx = mech_ctx._plugin_context
        ctx.session.begin = mock.MagicMock()
        mock_super.returnValue = interface_info
        mock_core_plugin.side_effect = mock.MagicMock()
        mock_network_info.side_effect = self.side_effect_network_info
        ve_map = {'ve 402': '12.0.0.1/24'}
        vlan_map.update({'402': ve_map})
        config_map.update({'2.2.2.2': vlan_map})
        self.driver.remove_router_interface(ctx, ROUTER,
                                            interface_info)
        self.assertDictSupersetOf(config_map, delete_config_map)

    def side_effect(self, dev_name):
        """
        Mock _get_driver method and return FakeDriver
        """
        device = device_map.get(dev_name)
        return FakeDriver(device)

    def side_effect_error(self, dev_name):
        """
        Mock _get_driver method and return FakeDriver
        """
        device = device_map.get(dev_name)
        return FakeDriver(device, error=True)

    def _get_network_context(self, physnet, network_type):
        """
        Create mock network context
        """
        network = {
            'id': 1,
            'name': 'private',
            'tenant_id': 1,
            'vlan': 200,
            'network_type': network_type,
            'provider:segmentation_id': 200
        }

        network_segments = [{
            'id': 1,
            'segmentation_id': 200,
            'network_type': network_type,
            'physical_network': physnet
        }]

        _plugin_context = context.get_admin_context()
        return FakeNetworkContext(network, network_segments, _plugin_context)

    def side_effect_network_info(self, vlan, ip):
        """
        Mock _get_driver method and return FakeDriver
        """

        return vlan_id, gateway_ip_cidr


class FakeNetworkContext(object):

    """To generate network context for testing purposes only."""

    def __init__(self, network, segments=None, original_network=None):
        self._network = network
        self._original_network = original_network
        self._segments = segments

    @property
    def current(self):
        return self._network

    @property
    def _plugin_context(self):
        return self._original_network

    @property
    def network_segments(self):
        return self._segments


class FakeDriver(object):

    """
    Fake driver which will implement create and delete
    network. Create network will update the global dictionary with
    the address of the device along with vlan and ports to be tagged.
    Example : {'10.10.23.1':{'200':['1/1/1', '1/1/2']}}

    Delete network will delete the corresponding entry from the dictionary.
    """

    def __init__(self, device, error=None):
        self.error = error
        self.device = device
        self.address = device.get('address')

    def add_router_interface(self, vlan_id, gateway_ip_cidr):
        if self.error is INVALID_INPUT:
            raise Exception("Ethernet Driver : Create"
                            "network failed: error= Invalid Input")
        elif self.error is RANGE_ERROR:
            raise Exception("Configuring router interface failed: "
                            "ve out of range error")
        elif self.error:
            raise Exception("Add Router Interface failed")

        interface_map.update({VE + vlan_id: gateway_ip_cidr})
        vlan_map.update({vlan_id: interface_map})
        config_map.update({self.address: vlan_map})

    def remove_router_interface(self, vlan_id):
        if vlan_id in config_map[self.address]:
            config_map[self.address].update({vlan_id: {}})
        else:
            raise Exception("vlan is not there")



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from sqlalchemy.orm import exc as orm_exc

from neutron.db import model_base
from neutron.db import models_v2

from networking_brocade.vyatta.common import config as vyatta_config
from networking_brocade.vyatta.common import exceptions as v_exc


@six.add_metaclass(abc.ABCMeta)
class FetcherAbstract(object):
    value = _unset = object()

    def __call__(self, context):
        if self.value is self._unset:
            self.value = self._lookup(context)
            if isinstance(self.value, model_base.BASEV2):
                context.session.expunge(self.value)
        return self.value

    def reset(self):
        try:
            del self.value
        except AttributeError:
            pass

    @abc.abstractmethod
    def _lookup(self, context):
        pass


class ManageNetworkFetcher(FetcherAbstract):
    def _lookup(self, context):
        q = context.session.query(models_v2.Network)
        q = q.filter_by(
            id=vyatta_config.VROUTER.management_network_id)
        try:
            value = q.one()
        except orm_exc.NoResultFound:
            raise v_exc.InvalidParameter(
                cause=_('Management network {0} specified in vRouter plugin '
                        'configuration file does not exist').format(
                    vyatta_config.VROUTER.management_network_id))
        return value


get_management_network = ManageNetworkFetcher()



# Copyright 2015 Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

if sys.version_info >= (2, 7):
    import unittest
else:
    import unittest2 as unittest

MLX_TEST_BASE_PATH = './networking_brocade/mlx/tests'


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    base_path = os.path.split(os.path.dirname(os.path.abspath(__file__)))[0]
    base_path = os.path.split(base_path)[0]
    test_dirs = {'./networking_brocade/tests',
                 './networking_brocade/vdx/tests/unit/ml2/drivers/brocade',
                 MLX_TEST_BASE_PATH + '/unit/ml2/drivers/brocade',
                 MLX_TEST_BASE_PATH + '/unit/services/l3_router/brocade',
                 './networking_brocade/vyatta/tests'
                 }
    for test_dir in test_dirs:
        if not pattern:
            suite.addTests(loader.discover(test_dir, top_level_dir=base_path))
        else:
            suite.addTests(loader.discover(test_dir, pattern=pattern,
                           top_level_dir=base_path))
    return suite



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Constants used by device_connector
"""

ENABLE_TERMINAL_CMD = "en\r"
CONFIGURE_TERMINAL_CMD = "conf t\r"
CONFIGURE_VLAN = "vlan {vlan_id}\r"
CONFIGURE_ETHERNET = "tagged ethe {port_number}\r"
DELETE_CONFIGURED_VLAN = "no vlan {vlan_id}\r"
EXIT = "exit\r"
SHOW_VERSION = "show version\r"
CTRL_C = '\x03'

"""
Constants used by ni_driver
"""
CONFIGURE_ROUTER_INTERFACE = "router-interface ve {vlan_id}\r"
CONFIGURE_INTERFACE = "interface ve {vlan_id}\r"
CONFIGURE_GATEWAY_IP = "ip address {gateway_ip_addr}\r"
DELETE_ROUTER_INTERFACE = "no router-interface ve {vlan_id}\r"



# Copyright 2015 Brocade Communications Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Common driver for both FI and NI devices
"""

from oslo_log import log as logging
from oslo_utils import importutils

CONNECTION_FACTORY = ("networking_brocade.mlx.ml2.connector_factory."
                      "ConnectorFactory")
INVALID_INPUT = "Invalid input"
RANGE_ERROR = "outside of allowed max"
ERROR = "Error"
DUPLICATE_IP = "Error: Duplicate ip address"

LOG = logging.getLogger(__name__)


class FiNiDriver(object):

    """
    Base driver which handles all common CLI Commands for both FI and NI
    devices
    """

    def __init__(self, device):
        self.device = device

    def create_network(self, vlan_id, ports):
        """
        Invoked when user creates a VLAN network. Triggers creation of
        VLAN on the devices listed in the configuration template

        :param:vlan_id: vlan id to add
        :param:ports: ports to be tagged to the above vlan id
        :raises: Exception
        """

        connector = importutils.import_object(
            CONNECTION_FACTORY).get_connector(self.device)
        connector.connect()
        response = connector.create_vlan(vlan_id, ports)
        msg = self.get_error_msg(response)
        if msg:
            raise Exception(reason=msg)
        connector.close_session()

    def delete_network(self, vlan_id):
        """
        Invoked when user deletes a VLAN network. Triggers deletion of
        VLAN on the devices listed in the configuration template

        :param:vlan_id: vlan id to delete
        """

        connector = importutils.import_object(
            CONNECTION_FACTORY).get_connector(self.device)
        connector.connect()
        connector.delete_vlan(vlan_id)
        connector.close_session()

    def add_router_interface(self, vlan_id, gateway_ip_cidr):
        """
        Invoked when user creates router interface. Triggers creation of
        router interface based on the VLAN ID.

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        :param:gateway_ip_cidr: Gateway IP address with network length.
        :raises: Exception
        """
        connector = importutils.import_object(
            CONNECTION_FACTORY).get_connector(self.device)
        connector.connect()
        response = connector.create_l3_router(vlan_id, gateway_ip_cidr)
        msg = self.get_error_msg(response)
        if msg:
            raise Exception(reason=msg)
        connector.close_session()

    def remove_router_interface(self, vlan_id):
        """
        Invoked when user deletes router interface. Triggers deletion of
        router interface based on the VLAN ID.

        :param:vlan_id: vlan id to which the router interface is part of
        :raises: Exception
        """
        connector = importutils.import_object(
            CONNECTION_FACTORY).get_connector(self.device)
        connector.connect()
        response = connector.delete_l3_router(vlan_id)
        msg = self.get_error_msg(response)
        if msg:
            raise Exception(reason=msg)
        connector.close_session()

    def get_error_msg(self, response_list):
        """
        Checks for error in response

        :param:response_list: List of response from device
        :returns:msg - Message will contain error description in case of an
            error, None otherwise.
        """
        msg = None
        for response in response_list:
            if INVALID_INPUT in response:
                msg = _("Ethernet Driver : Create"
                        "network failed: error= Invalid Input")
                LOG.error(msg)
                break
            elif RANGE_ERROR in response_list:
                msg = _("Configuring router interface failed: "
                        "ve out of range error")
                LOG.error(msg)
                break
            elif ERROR in response_list:
                msg = _("Configuring router interface failed: "
                        "vlan not associated to router interface")
                LOG.error(msg)
                break
        return msg
